# ISC License

# Copyright (c) 2020-2023, Hunter WB <hunterwb.com>

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Unicode to ASCII transliteration"""

from sys import intern
from zlib import MAX_WBITS, decompress
from pathlib import Path

def read_binary(resource):
    data_dir = Path(__file__).parent / "_data"
    return (data_dir / resource).read_bytes()

_blocks = {}


def anyascii(string):
    # type: (str) -> str
    """Transliterate a string into ASCII."""
    try:
        if string.isascii():
            return string
    except AttributeError:
        pass
    result = []
    for char in string:
        codepoint = ord(char)
        if codepoint <= 0x7F:
            result.append(char)
            continue
        blocknum = codepoint >> 8
        lo = codepoint & 0xFF
        try:
            block = _blocks[blocknum]
        except KeyError:
            try:
                b = read_binary("%03x" % blocknum)
                s = decompress(b, -MAX_WBITS).decode("ascii")
                block = tuple(map(intern, s.split("\t")))
            except FileNotFoundError:
                block = ()
            _blocks[blocknum] = block
        if len(block) > lo:
            result.append(block[lo])
    return "".join(result)



import torch
from torch import nn
import os
import json
import torch

class WavEncoder(nn.Module):
    """Handles only the encoding part of WavTokenizer"""
    
    def __init__(self, feature_extractor):
        super().__init__()
        self.feature_extractor = feature_extractor
        
    def _create_config(self):
        config = {
            'feature_extractor_class': 'decoder.feature_extractors.EncodecFeatures',
            'feature_extractor_config': {
                'encodec_model': 'encodec_24khz',
                'bandwidths': [6.6, 6.6, 6.6, 6.6],
                'train_codebooks': True,
                'num_quantizers': 1,
                'dowmsamples': [8, 5, 4, 2],
                'vq_bins': 4096,
                'vq_kmeans': 200
            }
        }
        return config
    
    def save_pretrained(self, save_directory):
        """Save the model and its configuration to a directory"""
        os.makedirs(save_directory, exist_ok=True)
        
        model_path = os.path.join(save_directory, "encoder_model.pt")
        torch.save(self.state_dict(), model_path)
        
        config = self._create_config()
        config_path = os.path.join(save_directory, "config.json")
        with open(config_path, 'w') as f:
            json.dump(config, f, indent=2)
    
    @classmethod
    def from_pretrained(cls, model_directory):
        """Load a model and its configuration from a directory"""
        config_path = os.path.join(model_directory, "config.json")
        with open(config_path, 'r') as f:
            config = json.load(f)
        
        # Import feature extractor class
        module_name, class_name = config['feature_extractor_class'].rsplit('.', 1)
        module = __import__(module_name, fromlist=[class_name])
        feature_extractor_cls = getattr(module, class_name)
        
        # Create feature extractor
        feature_extractor = feature_extractor_cls(**config['feature_extractor_config'])
        
        model = cls(feature_extractor)
        
        model_path = os.path.join(model_directory, "encoder_model.pt")
        state_dict = torch.load(model_path, map_location='cpu')
        model.load_state_dict(state_dict)
        
        return model

    @torch.inference_mode()
    def forward(self, audio_input, **kwargs):
        """Encode audio into discrete codes"""
        features, discrete_codes, _ = self.feature_extractor.infer(audio_input, **kwargs)
        return features, discrete_codes

class WavDecoder(nn.Module):
    """Handles only the decoding part of WavTokenizer"""
    
    def __init__(self, backbone, head, codebook_weights):
        super().__init__()
        self.backbone = backbone
        self.head = head
        self.register_buffer('codebook_weights', codebook_weights)
    
    def _create_config(self):
        config = {
            'backbone_class': 'decoder.models.VocosBackbone',
            'head_class': 'decoder.heads.ISTFTHead',
            'backbone_config': {
                'input_channels': 512,
                'dim': 768,
                'intermediate_dim': 2304,
                'num_layers': 12,
                'adanorm_num_embeddings': 4
            },
            'head_config': {
                'dim': 768,
                'n_fft': 1280,
                'hop_length': 320,
                'padding': 'same'
            }
        }
        return config
    
    def save_pretrained(self, save_directory):
        """Save the model and its configuration to a directory"""
        os.makedirs(save_directory, exist_ok=True)
        
        model_path = os.path.join(save_directory, "decoder_model.pt")
        save_dict = {
            'model_state_dict': self.state_dict(),
            'codebook_weights': self.codebook_weights
        }
        torch.save(save_dict, model_path)
        
        config = self._create_config()
        config_path = os.path.join(save_directory, "config.json")
        with open(config_path, 'w') as f:
            json.dump(config, f, indent=2)
    
    @classmethod
    def from_pretrained(cls, model_directory):
        """Load a model and its configuration from a directory"""
        config_path = os.path.join(model_directory, "config.json")
        with open(config_path, 'r') as f:
            config = json.load(f)
        
        # Import and create backbone
        module_name, class_name = config['backbone_class'].rsplit('.', 1)
        module = __import__(module_name, fromlist=[class_name])
        backbone_cls = getattr(module, class_name)
        backbone = backbone_cls(**config['backbone_config'])
        
        # Import and create head
        module_name, class_name = config['head_class'].rsplit('.', 1)
        module = __import__(module_name, fromlist=[class_name])
        head_cls = getattr(module, class_name)
        head = head_cls(**config['head_config'])
       
        model_path = os.path.join(model_directory, "decoder_model.pt")
        checkpoint = torch.load(model_path, map_location='cpu')
        
        model = cls(backbone, head, checkpoint['codebook_weights'])
        model.load_state_dict(checkpoint['model_state_dict'])
        
        return model

    def codes_to_features(self, codes):
        """Convert discrete codes to features using codebook"""
        if codes.dim() == 2:
            codes = codes.unsqueeze(1)
            
        n_bins = self.codebook_weights.size(0) // len(codes)
        offsets = torch.arange(0, n_bins * len(codes), n_bins, device=codes.device)
        embeddings_idxs = codes + offsets.view(-1, 1, 1)
        
        features = torch.nn.functional.embedding(embeddings_idxs, self.codebook_weights).sum(dim=0)
        features = features.transpose(1, 2)
        
        return features

    @torch.inference_mode()
    def forward(self, features_input, **kwargs):
        """Decode features to audio"""
        x = self.backbone(features_input, **kwargs)
        audio_output = self.head(x)
        return audio_output


from loguru import logger
from tqdm import tqdm
from packaging import version

from .info import GenerationType
from .config import GenerationConfig

try:
    from llama_cpp import Llama, llama_token_is_eog
    from llama_cpp import __version__ as llama_cpp_version
    _GGUF_AVAILABLE = True
except:
    llama_cpp_version = "0.0.0"
    _GGUF_AVAILABLE = False
    raise ImportError(
        "llama.cpp Python bindings not found. This is required for GGUF model support.\n\n"
        "To install, please follow our installation guide:\n"
        "https://github.com/edwko/OuteTTS?tab=readme-ov-file#installation\n\n"
    )

CURRENT_VERSION = version.parse(llama_cpp_version)
VERSION_0_3_7 = version.parse("0.3.7")

class GGUFModel:
    def __init__(
            self,
            model_path: str,
            n_gpu_layers: int = 0,
            max_seq_length: int = 4096,
            additional_model_config: dict = {}
    ) -> None:

        if not _GGUF_AVAILABLE:
            raise ImportError(
                "llama_cpp python module not found."
            )

        additional_model_config["n_ctx"] = max_seq_length
        self.model = Llama(
            model_path=model_path,
            n_gpu_layers=n_gpu_layers,
            last_n_tokens_size=64,
            **additional_model_config
        )

    def is_eog(self):
        if CURRENT_VERSION >= VERSION_0_3_7:
            return self.model._model.vocab
        else:
            return self.model._model.model

    def generate(self, input_ids: list[int], config: GenerationConfig):
        if config.generation_type == GenerationType.STREAM:
            return self._generate_stream(input_ids, config)
        return self._generate(input_ids, config)

    def _generate_stream(self, input_ids: list[int], config: GenerationConfig):
        input_size = len(input_ids)
        gen = tqdm(self.model.generate(
            input_ids,
            temp=config.sampler_config.temperature,
            repeat_penalty=config.sampler_config.repetition_penalty,
            top_k=config.sampler_config.top_k,
            top_p=config.sampler_config.top_p,
            min_p=config.sampler_config.min_p,
            mirostat_eta=config.sampler_config.mirostat_eta,
            mirostat_tau=config.sampler_config.mirostat_tau,
            **config.additional_gen_config,
        ))
        for token in gen:
            yield token
            input_size += 1
            if (llama_token_is_eog(self.is_eog(), token) or 
                input_size >= config.max_length):
                break
            gen.set_postfix({"tokens": input_size,  "max tokens": config.max_length})

    def _generate(self, input_ids: list[int], config: GenerationConfig) -> list:
        new_tokens = []
        for token in self._generate_stream(input_ids, config):
            new_tokens.append(token)
        return new_tokens


from loguru import logger
from tqdm import tqdm
from packaging import version
import json
import requests
import time
from tqdm.auto import tqdm as atqdm
import asyncio
import aiohttp
from urllib.parse import urljoin

from .info import GenerationType
from .config import GenerationConfig

class LlamaCPPServerModel:
    def __init__(
            self,
            model_path: str,
            n_gpu_layers: int = 0,
            max_seq_length: int = 4096,
            additional_model_config: dict = {},
            tokenizer = None
    ) -> None:
        self.tokenizer = tokenizer
        
    def generate(self, input_ids: str, config: GenerationConfig):
        if config.generation_type == GenerationType.STREAM:
            return self._generate_stream(input_ids, config)
        return self._generate(input_ids, config)

    def _generate_stream(self, input_ids: str, config: GenerationConfig):

        requests_config = {
            "top_p": config.sampler_config.top_p,
            "top_k": config.sampler_config.top_k,
            "min_p": config.sampler_config.top_p,
            "stream": True,
            "repeat_penalty": config.sampler_config.repetition_penalty,
            "temperature": config.sampler_config.temperature,
            "stop": ["<|im_end|>"],
            "n_predict": config.max_length
        }
        requests_config["prompt"] = input_ids
        try:
            url = urljoin(config.server_host, "/completion")
            response = requests.post(
                url,
                json=requests_config,
                stream=True
            )
            gen = tqdm(response.iter_lines())
            for line in gen:
                if line:
                    try:
                        text = json.loads(line.decode('utf-8').replace('data: ', ''))['content']
                        if not text:
                            time.sleep(0.001)
                            continue
                        yield text
                    except Exception as e:
                        logger.warning(e)
        except Exception as e:
            logger.error(e)

    def _generate(self, input_ids: str, config: GenerationConfig) -> list:
        new_tokens = []
        for token in self._generate_stream(input_ids, config):
            new_tokens.append(token)
        return self.tokenizer.encode("".join(new_tokens), add_special_tokens=False)
    
class LlamaCPPServerAsyncModel:
    def __init__(
            self,
            model_path: str,
            n_gpu_layers: int = 0,
            max_seq_length: int = 4096,
            additional_model_config: dict = {}
    ) -> None:
        self.loop = asyncio.new_event_loop()
        self._session: aiohttp.ClientSession | None = None
        self.sem = None

    async def _ensure_session(self):
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession()

    async def _generate(
        self,
        prompt: str,
        request_id: int,
        config: GenerationConfig,
        pbar: atqdm,
        lock: asyncio.Lock,
    ):

        async with self.sem:

            logger.success(f"started: {request_id}")

            payload = {
                "top_p": config.sampler_config.top_p,
                "top_k": config.sampler_config.top_k,
                "min_p": config.sampler_config.top_p,
                "repeat_penalty": config.sampler_config.repetition_penalty,
                "temperature": config.sampler_config.temperature,
                "stop": ["<|im_end|>"],
                "n_predict": config.max_length,
                "stream": True,
                "prompt": prompt,
            }

            final_output = ""

            try:
                url = urljoin(config.server_host, "/completion")
                async with self._session.post(
                    url, json=payload
                ) as resp:
                    resp.raise_for_status()
                    async for raw in resp.content:
                        for line in raw.split(b"\n"):
                            if not line.strip():
                                continue
                            try:
                                text = json.loads(
                                    line.decode("utf-8").removeprefix("data: ")
                                )["content"]
                                if not text:
                                    # throttle tiny empty events
                                    await asyncio.sleep(0.001)
                                    continue

                                final_output += text

                                async with lock:
                                    pbar.update(1)
                                    speed = pbar.format_dict.get("rate", 0)
                                    rtf = round(1 / (speed / 150), 4) if speed else 0
                                    pbar.set_postfix({"RTF": rtf})
                            except Exception as e:
                                logger.warning(f"[{request_id}] parse error: {e}")
            except Exception as e:
                logger.error(f"[{request_id}] request failed: {e}")

            print("\n")
            logger.success(f"Batch {request_id} finished.\n")

            return {"text": final_output, "id": request_id}
    
    async def _generate_batch_async(self, input_ids: list[str], config: GenerationConfig):
        """Helper coroutine to run multiple _generate tasks in parallel and gather results in order."""
        pbar = atqdm(total=None)
        lock = asyncio.Lock()
        self.sem = asyncio.Semaphore(config.max_batch_size)

        await self._ensure_session()

        tasks = [
            asyncio.create_task(
                self._generate(i, idx, config, pbar, lock)
            )
            for idx, i in enumerate(input_ids)
        ]
        return await asyncio.gather(*tasks)
    
    def generate_batch(self, input_ids: list[str], config: GenerationConfig):
        try:
            output = self.loop.run_until_complete(
                self._generate_batch_async(input_ids, config)
            )
            return output
        finally:
            # clean up session
            if self._session and not self._session.closed:
                self.loop.run_until_complete(self._session.close())


import whisper
from loguru import logger

def transcribe_once(audio_path: str, model: str = "turbo", device = None, language: str = None):
    logger.info(f"Loading model {model}")
    model = whisper.load_model(model, device=device) 
    logger.info(f"Transcribing {audio_path}")
    text = model.transcribe(audio_path, language=language)["text"]
    logger.success(f"Transcription: {text}")
    return text

def transcribe_once_word_level(audio_path: str, model: str = "turbo", device = None, language: str = None):
    logger.info(f"Loading model {model}")
    model = whisper.load_model(model, device=device) 
    logger.info(f"Transcribing {audio_path}")
    text = model.transcribe(audio_path, word_timestamps=True, language=language)
    logger.success(f"Transcription: {text['text'].strip()}")
    return text


import re
import uroman as ur
import MeCab
import string
import unicodedata

def check_language(text: str) -> bool:
    return any('\u3040' <= char <= '\u309f' or
                '\u30a0' <= char <= '\u30ff' or
                '\u4e00' <= char <= '\u9fff' or
                '\u4e00' <= char <= '\u9fff' and 
                any('\u3040' <= c <= '\u309f' for c in text)
                for char in text)

def get_punctuations():
    ascii_punctuation = set(string.punctuation)
    punctuation_categories = {'Pc', 'Pd', 'Pe', 'Pf', 'Pi', 'Po', 'Ps'}
    unicode_punct = set()
    
    # Check characters in common Unicode ranges
    for i in range(0x10000):  # First 65536 Unicode code points
        char = chr(i)
        category = unicodedata.category(char)
        if category in punctuation_categories:
            unicode_punct.add(char)
    
    all_punctuation = ascii_punctuation.union(unicode_punct)
    return sorted(all_punctuation)

def normalize_token_spacing(text: str):
    text = text.replace("'", "").replace("`", "")
    WHITESPACE_PATTERN = re.compile(r'(?<!\s)\s(?!\s)')
    text =  MeCab.Tagger("-Owakati").parse(text).split()
    text = [i for i in text if i not in ['', ' ']]
    result = text[0]  
    punctuations = get_punctuations()
    for token in text[1:]:
        if token not in punctuations:
            result += '  '
        else:
            result += ' '
        result += token
    return WHITESPACE_PATTERN.sub('', result)

def text_normalizations(text: str) -> str:
    # Normalize whitespace characters (newlines, tabs, etc.) to single spaces
    text = re.sub(r'\s+', ' ', text)
    text = text.replace("…", "...")  # Replace ellipsis character with three dots
    
    # Strip leading/trailing whitespace
    text = text.strip()
    
    # Normalize common Unicode characters to ASCII equivalents
    text = re.sub(r'[“”]', '"', text)   # Curly quotes to straight quotes
    text = re.sub(r'[‘’]', "'", text)   # Curly single quotes
    text = re.sub(r'[–—]', '-', text)   # Various dashes to hyphen
    
    # Remove control characters
    text = re.sub(r'[\x00-\x1F\x7F-\x9F]', '', text)
    
    return text

def get_words(text: str):
    text = text_normalizations(text)
    text = normalize_token_spacing(text)
    if check_language(text):
        mecab = MeCab.Tagger("-Owakati")
        return mecab.parse(text).strip().split()
    return text.split()



import os

os.environ["HF_HUB_DISABLE_TELEMETRY"] = "1"

__version__ = "0.4.4"

from .interface import Interface
from .models.info import Backend, InterfaceVersion, Models, LlamaCppQuantization, GenerationType
from .models.config import ModelConfig, GenerationConfig, SamplerConfig



from transformers import AutoTokenizer
import re
import unicodedata
import ftfy

from .tokens import SpecialTokens

class PromptProcessor:
    def __init__(self, tokenizer_path: str):
        self.special_tokens = SpecialTokens()

        if tokenizer_path:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)

            self.c1 = {}
            self.c2 = {}
            self.get_audio_token_map()

        self.input_prompt = "{bos}\n{text_start}{text}{text_end}\n{audio_start}\n"
        self.global_features = "{fs}{codes}{fe}\n"

    def get_audio_token_map(self):
        self.c1 = {
            self.tokenizer.encode(self.special_tokens.c1.format(i), add_special_tokens=False)[0]: i
            for i in range(1025)
        }
        self.c2 = {
            self.tokenizer.encode(self.special_tokens.c2.format(i), add_special_tokens=False)[0]: i
            for i in range(1025)
        }

    def get_features(self, f: dict):
        features = {
            "energy": f.get("energy", 0),
            "spectral_centroid": f.get("spectral_centroid", 0),
            "pitch": f.get("pitch", 0)
        }
        return [f"<|{k}_{v}|>" for k,v in features.items()]
    
    def get_global_features(self, f: dict):
        return self.global_features.format(
            fs=self.special_tokens.global_features_start,
            codes="".join(self.get_features(f)),
            fe=self.special_tokens.global_features_end
        )

    def create_codes(self, words: dict):
        codes = []
        for i in words:
            word = i['word'] + self.special_tokens.features + self.special_tokens.time.format(i['duration'])
            word += "".join(self.get_features(i['features']))
            pairs = []

            for idx in range(len(i['c1'])):
                c1 = self.special_tokens.c1.format(i['c1'][idx])
                c2 = self.special_tokens.c2.format(i['c2'][idx])
                pairs.append(
                    f"{c1}{c2}"
                )

            word += self.special_tokens.code + "".join(pairs) 
            codes.append(self.special_tokens.word_start + word + self.special_tokens.word_end)

        return "\n".join(codes)

    def _init_prompt(self, text):
        return self.input_prompt.format(
            bos=self.special_tokens.bos,
            text_start=self.special_tokens.text_start,
            text=text,
            text_end=self.special_tokens.text_end,
            audio_start=self.special_tokens.audio_start
        )

    def _get_separator(self, text: str) -> str:
        has_hiragana = any('\u3040' <= c <= '\u309f' for c in text)
        has_katakana = any('\u30a0' <= c <= '\u30ff' for c in text)
        has_han = any('\u4e00' <= c <= '\u9fff' for c in text)
        has_hangul = any('\uac00' <= c <= '\ud7af' for c in text)
        
        if has_hiragana or has_katakana or has_han:
            return "。"
        elif has_hangul:
            return ". "
        else:
            return ". "
    
    def merge_speaker_text(self, input_text: str, speaker_text: str) -> str:
        speaker_text = speaker_text.strip()
        separator = self._get_separator(speaker_text)
        
        # Determine allowed endings based on the separator
        if separator == "。":
            allowed_ends = ['。', '？', '！', '?', '!']
        else:
            allowed_ends = ['.', '?', '!']

        rs = ""
        if speaker_text:
            last_char = speaker_text[-1]
            if last_char not in allowed_ends:
                rs = separator
            else:
                if separator != "。":
                    rs = " "

        output = speaker_text.strip() + rs + input_text.strip()
        
        return output, rs.strip()
    
    def text_normalizations(self, text: str) -> str:
        if not isinstance(text, str):
            # Or raise TypeError, depending on desired behavior
            return str(text) if text is not None else ""

        text = ftfy.fix_text(text)

        text = unicodedata.normalize('NFKC', text)

        text = text.replace("…", "...")
        text = re.sub(r'\.{2,}', '...', text)

        text = re.sub(r'[“”„‟«»]', '"', text)
        text = re.sub(r"[‘’‛‹›`´]", "'", text)

        text = re.sub(r'[–—―−‐]', '-', text)
        text = re.sub(r'-{2,}', '-', text)

        text = re.sub(r'[\x00-\x1F\x7F-\x9F\u00AD\u200B-\u200D\uFEFF]', '', text)

        text = re.sub(r'\s+', ' ', text)

        text = re.sub(r'\s+([,.?!:;])', r'\1', text)
        text = re.sub(r'([,.?!:;])(?=[^\s,.?!:;])', r'\1 ', text)
        text = re.sub(r"(\w)\s+'\s*(\w)", r"\1'\2", text) # e.g. "word ' s" -> "word's"
        text = re.sub(r"(\w)\s+'\s*([,.?!:;\s]|$)", r"\1'\2", text) # e.g. "word ' ." -> "word'."

        text = re.sub(r"(\w)\s*'\s*([tsdmreSNTDMRE])\b", r"\1'\2", text, flags=re.IGNORECASE) # e.g., "can 't" -> "can't"

        text = re.sub(r'([?!])\1+', r'\1', text) # !! -> !, ??? -> ?

        text = re.sub(r'\s+', ' ', text)
        text = text.replace('"',"").strip()

        return text

    def get_completion_prompt(self, text: str, speaker: dict = None):
        text = self.text_normalizations(text)
        
        if speaker is not None:
            text, separator = self.merge_speaker_text(text, speaker["text"])
            speaker["words"][-1]['word'] += separator
            codes = self.create_codes(speaker["words"])
        
        prompt = self._init_prompt(text)

        if speaker is not None:
            prompt += codes + "\n" + self.special_tokens.word_start

        return prompt

    def get_training_prompt(self, speaker: dict) -> str:
        text = self.text_normalizations(speaker["text"])
        words = speaker["words"]
        global_features = speaker["global_features"]

        prompt = self._init_prompt(text)
        prompt += self.get_global_features(global_features)
        prompt += self.create_codes(words)
        prompt += "\n" + self.special_tokens.audio_end + "\n" + self.special_tokens.eos + "\n"

        return prompt 

    def extract_audio_from_tokens(self, tokens: list[int]):
        codebook1 = [self.c1[i] for i in tokens if i in self.c1]
        codebook2 = [self.c2[i] for i in tokens if i in self.c2]
        t = min(len(codebook1), len(codebook2))
        codebook1 = codebook1[:t]
        codebook2 = codebook2[:t]
        return [codebook1, codebook2]





        
    
        



import sys
import os

base_path = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, base_path)
sys.path.insert(0, os.path.join(base_path, 'decoder'))
sys.path.insert(0, os.path.join(base_path, 'encoder'))


import re
import inflect
import uroman as ur
import MeCab
import string
from loguru import logger

from ...anyascii import anyascii
from .tokens import SpecialTokens

def clean_dashes(text: str) -> str:
    dashes = ['—', '–', '-']
    for dash in dashes:
        text = text.replace(dash, ' ')
    return text

class TextProcessor:
    def __init__(self):
        self.special_tokens = SpecialTokens()

        escaped_punct = [re.escape(p) for p in self.special_tokens.punctuation_tokens.keys()]
        self.PUNCT_PATTERN = re.compile(f"[^a-zA-Z\u0080-\uffff{''.join(escaped_punct)}\\s]")
        # Pre-compile regex patterns
        self.WHITESPACE_PATTERN = re.compile(r'(?<!\s)\s(?!\s)')
        self.MULTIPLE_SPACES_PATTERN = re.compile(r'\s+')
        self.NUMBER_PATTERN = re.compile(r'\d+(\.\d+)?')
        self.CLEAN_PATTERN = re.compile(r'[^a-z\s]')

        self.wakati = MeCab.Tagger("-Owakati")
        self.lec = inflect.engine()
        self.uroman = ur.Uroman()

    def join_punctuation(self, text: str, pun: dict):
        text = text.split()
        i = 0
        while i < len(text):
            if text[i] in pun and i > 0: 
                text[i-1] = text[i-1] + text[i] 
                text.pop(i) 
            else:
                i += 1
        # Remove standalone punctuation tokens that weren't successfully joined to words
        return [word for word in text if word not in pun and not all(c in string.punctuation for c in word)]
    
    def normalize_token_spacing(self, text: str):
        text = text.replace("'", "").replace("`", "")
        text = self.wakati.parse(text).split()
        text = [i for i in text if i not in ['', ' ']]
        result = text[0]  
        pun = self.special_tokens.punctuation_tokens
        for token in text[1:]:
            if token not in pun:
                result += '  '
            else:
                result += ' '
            result += token
        return self.WHITESPACE_PATTERN.sub('', result)
    
    def _process_text(self, text: str):
        text = self.normalize_token_spacing(text)
        text = anyascii(text)
        text = clean_dashes(text)
        text = self.NUMBER_PATTERN.sub(lambda x: self.lec.number_to_words(x.group()), text.lower())
        text = self.PUNCT_PATTERN.sub('', text)
        text = self.MULTIPLE_SPACES_PATTERN.sub(' ', text).strip().lower()
        original = text

        text = self.CLEAN_PATTERN.sub('', text)
        text = [{"word": i} for i in text.split()]
      
        pun = self.special_tokens.punctuation_tokens
        original_split = self.join_punctuation(original, pun)

        # Covers some edge cases
        for idx, orig_word in enumerate(original_split):
            processed_word = text[idx]['word']
            match = re.search(processed_word, orig_word.lower())
           
            if match:
                start, end = match.span()
                text[idx]["before"] = [pun[c] for c in orig_word[:start] if c in pun]
                text[idx]["after"] = [pun[c] for c in orig_word[end:] if c in pun]
            else:
                text[idx]["before"] = []
                text[idx]["after"] = []

        return text

    def process_text(self, text: str):
        chunks = re.split(r'(<\|emotion_start\|>.*?<\|emotion_end\|>)', text)
        final_result = []
        for chunk in chunks:
            if chunk.strip().startswith('<|emotion_start|>'):
                # Add emotion block unprocessed
                final_result.append({
                    'word': chunk,
                    'before': [],
                    'after': []
                })
            else:
                if chunk.strip():
                    processed = self._process_text(chunk)
                    final_result.extend(processed)
        return final_result

    def process_text_clean_only(self, text: str):
        text = self.normalize_token_spacing(text)
        text = anyascii(text)
        text = clean_dashes(text)
        text = self.NUMBER_PATTERN.sub(lambda x: self.lec.number_to_words(x.group()), text.lower())
        text = self.PUNCT_PATTERN.sub('', text)
        text = self.MULTIPLE_SPACES_PATTERN.sub(' ', text).strip().lower()
        text = self.CLEAN_PATTERN.sub('', text)
        return text

    def get_text(self, data: dict):
        temp = []
        for i in data:
            word = i["word"]
            if i.get("before",[]):
                word = "".join(i["before"]) + word
            if i.get("after", []):
                word += "".join(i["after"])
            temp.append(word)
        return self.special_tokens.space.join(temp)
    


import torch
from loguru import logger
from transformers import AutoModelForCausalLM, AutoTokenizer, TextIteratorStreamer
from .config import GenerationConfig
from .info import GenerationType
from transformers import LogitsProcessor
import transformers.generation.utils as generation_utils

# --------------------------------------------------------------------------------------------------------
# RepetitionPenaltyLogitsProcessor
# This patch implements a windowed repetition penalty for OuteTTS 1.0, 
# which is essential for maintaining high-quality speech synthesis output. 
# The standard HuggingFace implementation applies repetition penalties across the entire input context, 
# but OuteTTS requires a more focused approach that only considers the most recent tokens.
# --------------------------------------------------------------------------------------------------------

class RepetitionPenaltyLogitsProcessorPatch(LogitsProcessor):
    def __init__(self, penalty: float):
        penalty_last_n = 64
        logger.info(f"🔄 Using patched RepetitionPenaltyLogitsProcessor -> RepetitionPenaltyLogitsProcessorPatch | penalty_last_n: {penalty_last_n}")
        if penalty_last_n is not None:
            if not isinstance(penalty_last_n, int) or penalty_last_n < 0:
                raise ValueError(f"`penalty_last_n` has to be a non-negative integer, but is {penalty_last_n}")
        if not isinstance(penalty, float) or penalty <= 0:
            raise ValueError(f"`penalty` has to be a positive float, but is {penalty}")

        self.penalty_last_n = penalty_last_n
        self.penalty = penalty

    @torch.no_grad()
    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """
        Args:
            input_ids (`torch.LongTensor`):
                Indices of input sequence tokens in the vocabulary (shape `(batch_size, sequence_length)`).
            scores (`torch.FloatTensor`):
                Prediction scores of a language modeling head (shape `(batch_size, vocab_size)`).

        Returns:
            `torch.FloatTensor`: The modified prediction scores.
        """
        # Check if penalties should be applied
        if self.penalty_last_n == 0 or self.penalty == 1.0:
            return scores

        batch_size, seq_len = input_ids.shape
        vocab_size = scores.shape[-1]

        # Process each batch item independently
        for b in range(batch_size):
            # 1. Determine the penalty window
            start_index = max(0, seq_len - self.penalty_last_n)
            window_indices = input_ids[b, start_index:] # Shape: (window_len,)

            if window_indices.numel() == 0: # Skip if window is empty
                continue

            # 2. Find unique tokens within the window
            tokens_in_window = set(window_indices.tolist())

            # 3. Apply repetition penalty to the scores for this batch item
            for token_id in tokens_in_window:
                if token_id >= vocab_size:
                    continue 

                logit = scores[b, token_id]

                if logit <= 0:
                    logit *= self.penalty
                else:
                    logit /= self.penalty

                # Update the score
                scores[b, token_id] = logit

        return scores
    
generation_utils.RepetitionPenaltyLogitsProcessor = RepetitionPenaltyLogitsProcessorPatch
AutoModelForCausalLM.generate = generation_utils.GenerationMixin.generate

class HFModel:
    def __init__(
        self,
        model_path: str,
        device: str = None,
        dtype: torch.dtype = None,
        additional_model_config: dict = {}
    ) -> None:
        self.device = torch.device(
            device if device is not None
            else "cuda" if torch.cuda.is_available()
            else "cpu"
        )
        self.dtype = dtype
        self.model = AutoModelForCausalLM.from_pretrained(
            model_path,
            torch_dtype=dtype,
            **additional_model_config
        ).to(self.device)
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)

    def _generate(self, input_ids: torch.Tensor, config: GenerationConfig):
        if config.sampler_config.temperature > 0:
            config.additional_gen_config["do_sample"] = True
        return self.model.generate(
            input_ids,
            max_length=config.max_length,
            temperature=config.sampler_config.temperature,
            repetition_penalty=config.sampler_config.repetition_penalty,
            top_k=config.sampler_config.top_k,
            top_p=config.sampler_config.top_p,
            min_p=config.sampler_config.min_p,
            **config.additional_gen_config,
        )[0].tolist()

    def _generate_stream(self, input_ids: torch.Tensor, config: GenerationConfig):
        raise NotImplementedError("Stream generation is not supported for HF models.")

    def generate(self, input_ids: list[int], config: GenerationConfig):
        if config.generation_type == GenerationType.STREAM:
            return self._generate_stream(input_ids, config)
        return self._generate(input_ids, config)



# https://github.com/pytorch/audio

# BSD 2-Clause License

# Copyright (c) 2017 Facebook Inc. (Soumith Chintala), 
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Implimentation of CTC forced alignment
# https://pytorch.org/audio/stable/tutorials/ctc_forced_alignment_api_tutorial.html

import torchaudio
import torch
import torchaudio.functional as F
import inflect
import re
import uroman as ur
import MeCab

class CTCForcedAlignment:

    def __init__(self, device: str = None):
        self.device = torch.device(device if device is not None else "cuda" if torch.cuda.is_available() else "cpu")
        bundle = torchaudio.pipelines.MMS_FA
        self.sample_rate = bundle.sample_rate
        self.model = bundle.get_model(with_star=False).to(self.device)
        self.LABELS = bundle.get_labels(star=None)
        self.DICTIONARY = bundle.get_dict(star=None)
        self.lec = inflect.engine()
        self.uroman = ur.Uroman()
        self.wakati = MeCab.Tagger("-Owakati")
        
    def process_text(self, text: str):
        text = self.wakati.parse(text)
        text = self.uroman.romanize_string(text)
        text = re.sub(r'\d+(\.\d+)?', lambda x: self.lec.number_to_words(x.group()), text.lower())
        text = re.sub(r'[-_/,\.\\]', ' ', text)
        text = re.sub(r'[^a-z\s]', '', text)
        text = re.sub(r'\s+', ' ', text).strip()
        return text.split()
    
    def _unflatten(self, list_, lengths):
        assert len(list_) == sum(lengths)
        i = 0
        ret = []
        for l in lengths:
            ret.append(list_[i : i + l])
            i += l
        return ret

    def get_word(self, waveform, spans, num_frames, transcript):
        ratio = waveform.size(1) / num_frames
        x0 = int(ratio * spans[0].start)
        x1 = int(ratio * spans[-1].end)
        return {"x0": x0, "x1": x1, "word": transcript}
    
    def _extract_world_level(self, aligned_tokens, alignment_scores, transcript):
        token_spans = F.merge_tokens(aligned_tokens, alignment_scores)
        word_spans = self._unflatten(token_spans, [len(word) for word in transcript])
        return word_spans

    def _align(self, emission, tokens):
        targets = torch.tensor([tokens], dtype=torch.int32, device=torch.device("cpu"))
        alignments, scores = F.forced_align(emission.cpu(), targets, blank=0)
        alignments, scores = alignments[0], scores[0] 
        scores = scores.exp()  
        return alignments, scores

    def align(self, audio, transcript):
        waveform, sr = torchaudio.load(audio)
        if waveform.shape[0] > 1:
            waveform = waveform.mean(dim=0, keepdim=True)
        waveform = torchaudio.functional.resample(waveform, orig_freq=sr, new_freq=self.sample_rate)
        transcript = self.process_text(transcript)

        with torch.inference_mode():
            emission, _ = self.model(waveform.to(self.device))

        tokenized_transcript = [self.DICTIONARY[c] for word in transcript for c in word]
        alignments, scores = self._align(emission, tokenized_transcript)
        word_spans = self._extract_world_level(alignments, scores, transcript)
        num_frames = emission.size(1)

        outputs = [
            self.get_word(waveform, word_spans[i], num_frames, transcript[i]) 
            for i in range(len(word_spans))
        ]

        outputs[0]["x0"] = 0

        for i in range(len(outputs)):
            output = outputs[i]
            x0 = output["x0"]

            if i == len(outputs) - 1:
                x1 = output["x1"]
            else:
                x1 = outputs[i + 1]["x0"]

            outputs[i]["audio"] = waveform[:, x0:x1]

        return outputs
    
    def free(self):
        del self.model



from loguru import logger
import os

from .version.interface import (
    InterfaceHF, InterfaceLLAMACPP, 
    InterfaceEXL2, InterfaceEXL2Async,
    InterfaceVLLMBatch, 
    InterfaceLlamaCPPServer, InterfaceLlamaCPPServerAsyncBatch
)
from .models.config import ModelConfig
from .models.info import Backend, InterfaceVersion

def Interface(config: ModelConfig) -> (
        InterfaceHF | InterfaceLLAMACPP | 
        InterfaceEXL2 | InterfaceEXL2Async |
        InterfaceVLLMBatch |
        InterfaceLlamaCPPServer | InterfaceLlamaCPPServerAsyncBatch):

    if config.backend == Backend.HF:
        return InterfaceHF(config)
    elif config.backend == Backend.LLAMACPP:
        return InterfaceLLAMACPP(config)
    elif config.backend == Backend.EXL2:
        return InterfaceEXL2(config)
    elif config.backend == Backend.EXL2ASYNC:
        return InterfaceEXL2Async(config)
    elif config.backend == Backend.VLLM:
        warning_msg = "VLLM backend is experimental and may cause issues with audio generation."
        logger.warning(warning_msg)
        return InterfaceVLLMBatch(config)
    elif config.backend == Backend.LLAMACPP_SERVER:
        return InterfaceLlamaCPPServer(config)
    elif config.backend == Backend.LLAMACPP_ASYNC_SERVER:
        return InterfaceLlamaCPPServerAsyncBatch(config)

    raise ValueError(f"Invalid backend: {config.backend} - must be one of {list(Backend)}")



import dac
import torch
import tqdm 
import torch.nn as nn
import torchaudio
import math
import pyloudnorm as pyln
import numpy as np
from huggingface_hub import hf_hub_download
import os

from ..utils import helpers

def process_audio_tensor(
    audio: torch.Tensor,
    sample_rate: int = 24000,
    target_loudness: float = -18.0,
    peak_limit: float = -1,
    block_size: float = 0.400,
) -> torch.Tensor:
    device = audio.device
    dtype = audio.dtype
    audio_np = audio.detach().cpu().numpy()

    if audio_np.ndim > 1:
        if audio_np.shape[1] > 1:
            audio_np = np.mean(audio_np, axis=1)
        else:
            audio_np = np.squeeze(audio_np)

    original_length = len(audio_np)
    min_samples = int(block_size * sample_rate)
    
    if original_length < min_samples:
        pad_length = min_samples - original_length
        audio_padded = np.pad(audio_np, (0, pad_length), mode='constant')
    else:
        audio_padded = audio_np

    meter = pyln.Meter(sample_rate, block_size=block_size)
    measured_loudness = meter.integrated_loudness(audio_padded)
    normalized = pyln.normalize.loudness(audio_padded, measured_loudness, target_loudness)

    peak_value = np.max(np.abs(normalized))
    threshold_value = 10 ** (peak_limit / 20)
    if peak_value > threshold_value:
        normalized = pyln.normalize.peak(normalized, peak_limit)

    if original_length < min_samples:
        normalized = normalized[:original_length]

    normalized_tensor = torch.from_numpy(normalized).to(dtype=dtype).to(device).unsqueeze(0).unsqueeze(0)
    return normalized_tensor

class DacInterface:
    def __init__(self, device: str = None, model_path: str = None):
        
        if model_path is None:
            model_path = self._get_model()
            
        self.device = torch.device(device if device is not None else "cuda" if torch.cuda.is_available() else "cpu")
        self.model = dac.DAC.load(model_path).to(self.device).eval()
        self.sr = 24000

    def _get_model(self):
        cache = helpers.get_cache_dir()
        model_path = hf_hub_download(
            repo_id="ibm-research/DAC.speech.v1.0",
            filename="weights_24khz_1.5kbps_v1.0.pth",
            local_dir=os.path.join(cache, "dac"),
            local_files_only=False
        )
        return model_path

    def convert_audio(self, wav: torch.Tensor, sr: int, target_sr: int, target_channels: int):
        assert wav.dim() >= 2, "Audio tensor must have at least 2 dimensions"
        assert wav.shape[-2] in [1, 2], "Audio must be mono or stereo."
        *shape, channels, length = wav.shape
        if target_channels == 1:
            wav = wav.mean(-2, keepdim=True)
        elif target_channels == 2:
            wav = wav.expand(*shape, target_channels, length)
        elif channels == 1:
            wav = wav.expand(target_channels, -1)
        else:
            raise RuntimeError(f"Impossible to convert from {channels} to {target_channels}")
        wav = torchaudio.transforms.Resample(sr, target_sr)(wav)
        return wav
    
    def convert_audio_tensor(self, audio: torch.Tensor, sr):
        return self.convert_audio(audio, sr, self.sr, 1)
    
    def load_audio(self, path):
        wav, sr = torchaudio.load(path)
        return self.convert_audio_tensor(wav, sr).unsqueeze(0)
    
    def preprocess(self, audio_data):
        length = audio_data.shape[-1]
        right_pad = math.ceil(length / self.model.hop_length) * self.model.hop_length - length
        audio_data = nn.functional.pad(audio_data, (0, right_pad))
        return audio_data
    
    @torch.no_grad()
    def encode(self, x: torch.Tensor, win_duration: int = 5.0, verbose: bool = False):
        x = process_audio_tensor(x)
        device = x.device
        nb, nac, nt = x.shape
        x = x.reshape(nb * nac, 1, nt)
        n_samples = int(win_duration * self.sr)
        n_samples = int(math.ceil(n_samples / self.model.hop_length) * self.model.hop_length)
        hop = n_samples 
        codes = []
        range_fn = range if not verbose else tqdm.trange
        for i in range_fn(0, nt, hop):
            chunk = x[..., i:i + n_samples]
            audio_data = chunk.to(self.model.device)
            audio_data = self.preprocess(audio_data)
            _, c, _, _, _ = self.model.encode(audio_data, None)
            codes.append(c.to(device))
        codes = torch.cat(codes, dim=-1)
        return codes
    
    def apply_fade(self, audio):
        fade_in_sec  = 0.015
        fade_out_sec = 0.015

        total_len = audio.shape[-1]

        fade_in_len  = int(self.sr * fade_in_sec)
        fade_out_len = int(self.sr * fade_out_sec)

        max_fade = total_len // 2
        fade_in_len  = min(fade_in_len, max_fade)
        fade_out_len = min(fade_out_len, max_fade)

        fade_in  = torch.linspace(0., 1., fade_in_len,  device=audio.device)
        fade_out = torch.linspace(1., 0., fade_out_len, device=audio.device)

        audio[..., :fade_in_len]                         *= fade_in
        audio[..., total_len - fade_out_len : total_len] *= fade_out

        return audio

    @torch.no_grad()
    def decode(
        self,
        codes: torch.Tensor,
        verbose: bool = False,
        chunk_length: int = 2048
    ) -> torch.Tensor:
        model = self.model
        range_fn = range if not verbose else tqdm.trange
        original_device = codes.device
        recons = []
        for i in range_fn(0, codes.shape[-1], chunk_length):
            c = codes[..., i : i + chunk_length].to(model.device)
            z = model.quantizer.from_codes(c)[0]
            r = model.decode(z)
            recons.append(self.apply_fade(r.to(original_device)))
        recons = torch.cat(recons, dim=-1)
        return process_audio_tensor(recons)



import torch
import torchaudio
from loguru import logger
import numpy as np
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"

try: 
    import sounddevice as sd
except: 
    logger.warning("[playback] Failed to import sounddevice.")
try: 
    import pygame
except: 
    logger.warning("[playback] Failed to import pygame.")

class ModelOutput:
    def __init__(self, audio: torch.Tensor, og_sr: int):
        self.sr = 44100
        self.audio = self.resample(audio, og_sr, self.sr)

    def resample(self, audio: torch.Tensor, og_sr: int, to_sr: int):
        resampler = torchaudio.transforms.Resample(orig_freq=og_sr, new_freq=to_sr).to(audio.device)
        return resampler(audio)

    def save(self, path: str):
        if self.audio is None:
            logger.warning("Audio is empty, skipping save.")
            return
        
        audio_2d = self.audio.detach().cpu()
        if audio_2d.dim() == 1:
            audio_2d = audio_2d.unsqueeze(0)
        elif audio_2d.dim() > 2:
            audio_2d = audio_2d[0] if audio_2d.dim() == 3 else audio_2d[0, 0]
            if audio_2d.dim() == 1:
                audio_2d = audio_2d.unsqueeze(0)

        if not path.endswith(".wav"):
            path += ".wav"

        torchaudio.save(path, audio_2d, sample_rate=self.sr, encoding='PCM_S', bits_per_sample=16)
        logger.info(f"Saved audio to: {path}")

    def _sounddevice(self):
        try:
            sd.play(self.audio.flatten().detach().cpu().numpy(), self.sr)
            sd.wait()
        except Exception as e:
            logger.error(e)

    def _pygame(self):
        try:
            pygame.mixer.init(frequency=self.sr, channels=2)
            audio_data = self.audio[0].detach().cpu().numpy()
            sound_array = (audio_data * 32767).astype('int16')
            if sound_array.ndim == 1:
                sound_array = np.expand_dims(sound_array, axis=1)
                sound_array = np.repeat(sound_array, 2, axis=1)
            sound = pygame.sndarray.make_sound(sound_array)
            sound.play()
            pygame.time.wait(int(sound.get_length() * 1000))
            pygame.mixer.quit()
        except Exception as e:
            logger.error(e)

    def _invalid_backend(self):
        logger.warning(f"Invalid backend selected!")

    def play(self, backend: str = "sounddevice"):
        """
        backend: str -> "sounddevice", "pygame"
        """
        logger.warning("Playback might not always work reliably. Always verify by playing the saved file.")

        if self.audio is None:
            logger.warning("Audio is empty, skipping playback.")
            return

        backends = {
            "sounddevice": self._sounddevice,
            "pygame": self._pygame
        }

        backend = backend.lower()
        backends.get(backend, self._invalid_backend)()



from loguru import logger
import torch
from tqdm import tqdm
import asyncio
from tqdm.auto import tqdm as atqdm

from .config import GenerationConfig
from .info import GenerationType

try:
    from exllamav2 import ExLlamaV2, ExLlamaV2Config, ExLlamaV2Cache, ExLlamaV2Tokenizer
    from exllamav2.generator import (
        ExLlamaV2DynamicGenerator, ExLlamaV2DynamicJob, 
        ExLlamaV2Sampler, ExLlamaV2DynamicGeneratorAsync, 
        ExLlamaV2DynamicJobAsync
    )
    _EXL2_AVAILABLE = True
except Exception as e:
    _EXL2_AVAILABLE = False

class EXL2Model:
    def __init__(
            self,
            model_path: str,
            max_seq_length: int,
            additional_model_config: dict = {},
    ) -> None:

        if not _EXL2_AVAILABLE:
            raise ImportError(
                "exllamav2 python module not found."
                "To use the EXL2 model you must install exllamav2 manually."
            )
        
        config = ExLlamaV2Config(model_path)
        self.model = ExLlamaV2(config)
        self.cache = ExLlamaV2Cache(self.model, max_seq_len=1024*32, lazy=False)
        self.model.load_autosplit(self.cache, progress=True)
        self.tokenizer = ExLlamaV2Tokenizer(config)
        additional_dynamic_generator_config = additional_model_config.get("additional_dynamic_generator_config", {})
        self.generator = ExLlamaV2DynamicGenerator(
            model=self.model, cache=self.cache, tokenizer=self.tokenizer, **additional_dynamic_generator_config
        )

    def generate(self, input_ids: str, config: GenerationConfig):
        if config.generation_type == GenerationType.STREAM:
            return self._generate_stream(input_ids, config)
        return self._generate(input_ids, config)
    
    def _generate_stream(self, input_ids: str, config: GenerationConfig):
        gen_settings = ExLlamaV2Sampler.Settings(
            token_repetition_penalty=config.sampler_config.repetition_penalty,
            temperature=config.sampler_config.temperature,
            token_repetition_range=config.sampler_config.repetition_range,
            top_k=config.sampler_config.top_k,
            top_p=config.sampler_config.top_p,
            min_p=config.sampler_config.min_p,
            mirostat_eta=config.sampler_config.mirostat_eta,
            mirostat_tau=config.sampler_config.mirostat_tau,
            mirostat=config.sampler_config.mirostat,
            **config.additional_gen_config
        )
        tokens = self.tokenizer.encode(input_ids, add_bos=False, add_eos=False, encode_special_tokens=True)
        tokens_size = len(tokens)
        input_size = len(tokens.flatten().tolist())
        job = ExLlamaV2DynamicJob(
            input_ids=tokens,
            max_new_tokens=config.max_length - input_size,
            stop_conditions=[self.tokenizer.eos_token_id],
            gen_settings=gen_settings,
            identifier=1,
            decode_special_tokens=True,
        )
        self.generator.enqueue(job)
        with tqdm(desc="Generating") as pbar:
            while self.generator.num_remaining_jobs() > 0:
                results = self.generator.iterate()
                for result in results:
                    token = result.get("token_ids", None)
                    if token is None:
                        continue
                    pbar.update(1)
                    pbar.set_postfix({"tokens": input_size + tokens_size, "max tokens": config.max_length})
                    for t in token.flatten().tolist():
                        yield t
                        tokens_size += 1

    def _generate(self, input_ids: str, config: GenerationConfig):
        new_tokens = []
        for i in self._generate_stream(input_ids, config):
            new_tokens.append(i)
        return new_tokens
    
class EXL2ModelAsync:
    def __init__(
            self,
            model_path: str,
            max_seq_length: int,
            additional_model_config: dict = {},
    ) -> None:

        if not _EXL2_AVAILABLE:
            raise ImportError(
                "exllamav2 python module not found."
                "To use the EXL2 model you must install exllamav2 manually."
            )

        config = ExLlamaV2Config(model_path)
        self.model = ExLlamaV2(config)
        self.cache = ExLlamaV2Cache(self.model, max_seq_len=additional_model_config.get("exl2_cache_size", 8192*4), lazy=True)
        self.model.load_autosplit(self.cache, progress=True)
        self.tokenizer = ExLlamaV2Tokenizer(config)
        self.loop = asyncio.new_event_loop()
        additional_dynamic_generator_config = additional_model_config.get("additional_dynamic_generator_config", {})
        self.generator = self.loop.run_until_complete(self._create_generator(additional_dynamic_generator_config))
        self.sem = None
    
    async def _create_generator(self, additional_dynamic_generator_config):
        return ExLlamaV2DynamicGeneratorAsync(
            model=self.model, 
            cache=self.cache, 
            tokenizer=self.tokenizer, 
            **additional_dynamic_generator_config
        )

    async def _generate(
            self, 
            input_ids: str, 
            sampling_params, 
            request_id, 
            config: GenerationConfig,
            pbar: atqdm,
            lock: asyncio.Lock
        ):

        async with self.sem:

            tokens = self.tokenizer.encode(input_ids, add_bos=False, add_eos=False, encode_special_tokens=True)
            input_size = len(tokens.flatten().tolist())
            job = ExLlamaV2DynamicJobAsync(
                self.generator,
                input_ids=tokens,
                max_new_tokens=config.max_length - input_size,
                stop_conditions=[self.tokenizer.eos_token_id],
                gen_settings=sampling_params,
                identifier=request_id,
                decode_special_tokens=True,
            )
            text = ""

            async for result in job:
                delta_text = result.get("text")
                if delta_text is None:
                    await asyncio.sleep(0)
                    continue
                    
                text += delta_text

                async with lock:
                    pbar.update(result.get("token_ids").size(1))
                    speed = pbar.format_dict.get("rate", 0)
                    rtf = round(1/(speed/150), 4)
                    pbar.set_postfix({
                        "RTF": rtf,
                    })

            print("\n")
            logger.success(f"Batch {request_id} finished.\n")

            return {"text": text, "id": request_id}
        
    async def _generate_batch_async(self, input_ids: list[str], sampling_params, config: GenerationConfig):
        """Helper coroutine to run multiple _generate tasks in parallel and gather results in order."""
        pbar = atqdm(total=None)
        lock = asyncio.Lock()
        self.sem = asyncio.Semaphore(config.max_batch_size)

        tasks = [
            asyncio.create_task(
                self._generate(i, sampling_params, idx, config, pbar, lock)
            )
            for idx, i in enumerate(input_ids)
        ]
        return await asyncio.gather(*tasks)
    
    def generate_batch(self, input_ids: list[str], config: GenerationConfig):
        sampling_params = ExLlamaV2Sampler.Settings(
            token_repetition_penalty=config.sampler_config.repetition_penalty,
            temperature=config.sampler_config.temperature,
            token_repetition_range=config.sampler_config.repetition_range,
            top_k=config.sampler_config.top_k,
            top_p=config.sampler_config.top_p,
            min_p=config.sampler_config.min_p,
        )

        output = self.loop.run_until_complete(
            self._generate_batch_async(input_ids, sampling_params, config)
        )
        torch.cuda.empty_cache()
        return output


from transformers import AutoTokenizer
from loguru import logger
import re
import inflect
import uroman as ur
import MeCab

class PromptProcessor:
    def __init__(self, tokenizer_path: str):
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
        self.bos = "<|im_start|>"
        self.eos = "<|im_end|>"
        self.special_tokens = {
            "audio_code": "<|{}|>",
            "text_start": "<|text_start|>",
            "text_end": "<|text_end|>",
            "audio_start": "<|audio_start|>",
            "audio_end": "<|audio_end|>",
            "time": "<|t_{:.2f}|>",
            "code_start": "<|code_start|>",
            "code_end": "<|code_end|>",
            "text_sep": "<|text_sep|>"
        }
        self.text_prompt = "{bos}\n{text_start}{words}{text_end}\n{audio_start}\n"
        self.map_audio_tokens = self.get_audio_token_map()

        self.lec = inflect.engine()
        self.uroman = ur.Uroman()
        self.wakati = MeCab.Tagger("-Owakati")
        self.wakati_use = ["ja", "zh", "ko"]

    def get_audio_token_map(self) -> dict:
        return {
            self.tokenizer.encode(self.special_tokens["audio_code"].format(i), add_special_tokens=False)[0]: i
            for i in range(4100)
        }

    def process_text(self, text: str):
        text = self.wakati.parse(text)
        text = self.uroman.romanize_string(text)
        text = re.sub(r'\d+(\.\d+)?', lambda x: self.lec.number_to_words(x.group()), text.lower())
        text = re.sub(r'[-_/,\.\\]', ' ', text)
        text = re.sub(r'[^a-z\s]', '', text)
        text = re.sub(r'\s+', ' ', text).strip()
        return text.split()

    def create_audio_prompt(self, words: list) -> str:
        prompt = []
        for i in words:
            word = i["word"]
            duration = self.special_tokens["time"].format(i["duration"])
            tokens = "".join([self.special_tokens["audio_code"].format(c) for c in i["codes"]])
            prompt.append(f'{word}{duration}{self.special_tokens["code_start"]}{tokens}{self.special_tokens["code_end"]}')
        return "\n".join(prompt)
        
    def get_completion_prompt(self, text: str, speaker: dict = None) -> str:
        words = self.process_text(text)
        if speaker is not None:
            words = self.process_text(speaker["text"]) + words

        words = f"{self.special_tokens['text_sep']}".join([i.strip() for i in words])

        prompt = self.text_prompt.format(
            bos=self.bos, 
            text_start=self.special_tokens['text_start'], 
            words=words, 
            text_end=self.special_tokens['text_end'],
            audio_start=self.special_tokens['audio_start']
        )

        if speaker is not None:
            prompt += self.create_audio_prompt(speaker["words"])

        return prompt
    
    def get_training_prompt(self, text: str, speaker: dict) -> str:
        words = self.process_text(text)
        words = f"{self.special_tokens['text_sep']}".join([i.strip() for i in words])

        prompt = self.text_prompt.format(
            bos=self.bos, 
            text_start=self.special_tokens['text_start'], 
            words=words, 
            text_end=self.special_tokens['text_end'],
            audio_start=self.special_tokens['audio_start']
        )
        prompt += self.create_audio_prompt(speaker["words"])
        prompt += f"\n{self.special_tokens['audio_end']}\n{self.eos}\n"

        return prompt
    
    def extract_audio_from_tokens(self, tokens: list[int]) -> list[int]:
        return [self.map_audio_tokens[i] for i in tokens if i in self.map_audio_tokens]



import os
import platform

def get_cache_dir():
    cache_dir = os.path.join(
        os.getenv('APPDATA') if platform.system() == "Windows" 
        else os.path.expanduser("~/.cache"), 
        "outeai"
    )
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir


import re
from typing import List
import unicodedata
import MeCab

def check_language(text: str) -> bool:
    return any('\u3040' <= char <= '\u309f' or
                '\u30a0' <= char <= '\u30ff' or
                '\u4e00' <= char <= '\u9fff' or
                '\u4e00' <= char <= '\u9fff' and 
                any('\u3040' <= c <= '\u309f' for c in text)
                for char in text)

def tokenize_text(text: str) -> List[str]:
    """
    Tokenize text based on detected language.
    """
    if not text.strip():
        return []
    
    if check_language(text):
        mecab = MeCab.Tagger("-Owakati")
        return mecab.parse(text).strip().split()
    else:
        return text.split()

def count_words(text: str) -> int:
    """Count words in text using appropriate tokenization."""
    return len(tokenize_text(text))

def split_into_sentences(text: str) -> List[str]:
    """Split text into sentences using appropriate end markers."""
    # Common sentence end markers for various languages
    end_markers = r'[.!?。！？︕︖]+'
    
    # Split on sentence end markers while preserving them
    parts = re.split(f'({end_markers}\\s*)', text)
    
    sentences = []
    current_sentence = ""
    
    for i in range(0, len(parts), 2):
        current_sentence = parts[i]
        
        # Add the sentence end marker if it exists
        if i + 1 < len(parts):
            current_sentence += parts[i + 1]
        
        if current_sentence.strip():
            sentences.append(current_sentence.strip())
    
    return sentences

def chunk_text(text: str, min_words: int = 10, max_words: int = 30) -> List[str]:
    """
    Chunk multilingual text into segments with specified word count bounds.
    
    Args:
        text: Input text that may contain multiple languages
        min_words: Minimum words per chunk
        max_words: Maximum words per chunk
    
    Returns:
        List of text chunks
    """
    # Normalize unicode characters
    text = unicodedata.normalize('NFKC', text)

    # Remove excessive whitespace while preserving sentence boundaries
    text = re.sub(r'\s+', ' ', text).strip()
    
    if not text:
        return []
    
    sentences = split_into_sentences(text)
    chunks = []
    current_chunk = ""
    current_word_count = 0
    
    for sentence in sentences:
        sentence = sentence.strip()
        if not sentence:
            continue
            
        sentence_tokens = tokenize_text(sentence)
        sentence_word_count = len(sentence_tokens)
        
        # Handle long sentences
        if sentence_word_count > max_words:
            # Add current chunk if it exists
            if current_chunk:
                chunks.append(current_chunk)
                current_chunk = ""
                current_word_count = 0
            
            # Split long sentence into parts
            current_part = []
            word_count = 0
            
            for token in sentence_tokens:
                current_part.append(token)
                word_count += 1
                
                if word_count >= max_words:
                    if check_language(sentence):
                        chunks.append("".join(current_part))
                    else:
                        chunks.append(" ".join(current_part))
                    current_part = []
                    word_count = 0
            
            if current_part:
                if check_language(sentence):
                    chunks.append("".join(current_part))
                else:
                    chunks.append(" ".join(current_part))
            continue
        
        # Try to add sentence to current chunk
        if current_word_count + sentence_word_count <= max_words:
            if current_chunk:
                current_chunk += " " + sentence
            else:
                current_chunk = sentence
            current_word_count += sentence_word_count
        else:
            # Check if current chunk meets minimum requirement
            if current_word_count >= min_words:
                chunks.append(current_chunk)
                current_chunk = sentence
                current_word_count = sentence_word_count
            else:
                # Split the sentence to fit remaining space
                space_left = max_words - current_word_count
                current_part = sentence_tokens[:space_left]
                remaining_part = sentence_tokens[space_left:]
                
                if current_chunk:
                    if check_language(current_part):
                        chunks.append(current_chunk + "".join(current_part))
                    else:
                        chunks.append(current_chunk + " " + " ".join(current_part))
                else:
                    if check_language(current_part):
                        chunks.append("".join(current_part))
                    else:
                        chunks.append(" ".join(current_part))
                
                if check_language(remaining_part):
                    current_chunk = "".join(remaining_part)
                else:
                    current_chunk = " ".join(remaining_part)
                current_word_count = len(remaining_part)
    
    if current_chunk:
        chunks.append(current_chunk)
    
    return chunks



from dataclasses import dataclass, asdict, field
from typing import Dict

@dataclass
class SpecialTokens:
    """
    Dataclass containing special tokens used for text and audio processing.
    """
    bos: str = "<|im_start|>"
    eos: str = "<|im_end|>"
    audio_code: str = "<|{}|>"
    text_start: str = "<|text_start|>"
    text_end: str = "<|text_end|>"
    voice_characteristic_start: str = "<|voice_characteristic_start|>"
    voice_characteristic_end: str = "<|voice_characteristic_end|>"
    emotion_start: str = "<|emotion_start|>"
    emotion_end: str = "<|emotion_end|>"
    audio_start: str = "<|audio_start|>"
    audio_end: str = "<|audio_end|>"
    time: str = "<|t_{:.2f}|>"
    text_sep: str = "<|text_sep|>"
    space: str = "<|space|>"
    syllable_sep: str = "<|syllable_sep|>"

    punctuation_tokens: Dict[str, str] = field(default_factory=lambda: {
        # Basic English/Latin punctuation
        ".": "<|period|>",
        "!": "<|exclamation_mark|>",
        "?": "<|question_mark|>",
        ",": "<|comma|>",
        
        # Quotation marks across languages
        '"': "<|double_quote|>",
        "„": "<|low_double_quote|>",
        
        # Spanish/Italian specific
        "¡": "<|inverted_exclamation|>",
        "¿": "<|inverted_question|>",
        
        # Ellipsis
        "…": "<|ellipsis|>",
        "...": "<|ellipsis|>",
        
        # Chinese/Japanese sentence endings
        "。": "<|cjk_period|>",
        "！": "<|cjk_exclamation|>",
        "？": "<|cjk_question|>",
        "，": "<|cjk_comma|>",
        
        # Arabic sentence punctuation
        "؟": "<|arabic_question|>",
    })

    def to_dict(self) -> Dict[str, str]:
        """Convert the dataclass instance to a dictionary using asdict."""
        return asdict(self)


import io
import torch
import torch.nn.functional as F
from loguru import logger

from ...dac.interface import DacInterface
from ...whisper.transcribe import transcribe_once_word_level
from ...utils.preprocessing import text_normalizations

def calculate_pitch(
    audio_tensor: torch.Tensor,
    sr: int,
    min_freq: float = 75.0,
    max_freq: float = 600.0,
    frame_length: int = 400,
    hop_length: int = 160,
    threshold: float = 0.3
) -> torch.Tensor:
    """
    Calculate pitch frequencies for short audio clips using autocorrelation.
    
    Args:
        audio_tensor: Input audio tensor (1D or 2D [channels, samples])
        sr: Sampling rate
        min_freq: Minimum detectable frequency (Hz)
        max_freq: Maximum detectable frequency (Hz)
        frame_length: Analysis frame length in samples
        hop_length: Hop size in samples
        threshold: Voicing threshold (0.0-1.0)
    
    Returns:
        Tensor of pitch values (Hz) per frame
    """
    
    # Convert to mono and ensure 1D
    if audio_tensor.dim() > 1:
        audio_tensor = torch.mean(audio_tensor, dim=0)
    audio_tensor = audio_tensor.squeeze()
    
    # Pad audio to ensure full frame coverage
    num_samples = audio_tensor.shape[-1]
    pad_len = (frame_length - (num_samples % hop_length)) % hop_length
    audio_tensor = F.pad(audio_tensor, (0, pad_len))
    
    # Create frames using unfold
    frames = audio_tensor.unfold(0, frame_length, hop_length)
    num_frames = frames.shape[0]
    
    # Precompute window and move to device
    window = torch.hann_window(frame_length, device=audio_tensor.device)
    frames_windowed = frames * window
    
    # Compute autocorrelation using FFT
    f = torch.fft.rfft(frames_windowed, n=2*frame_length, dim=1)
    power_spectrum = f.real.pow(2) + f.imag.pow(2)
    autocorr = torch.fft.irfft(power_spectrum, dim=1)[:, :frame_length]
    
    # Find valid frequency range indices
    min_idx = max(1, int(sr / max_freq))
    max_idx = min(frame_length, int(sr / min_freq))
    
    # Find peak indices in valid range
    relevant_autocorr = autocorr[:, min_idx:max_idx]
    peak_values, peak_indices = torch.max(relevant_autocorr, dim=1)
    peak_indices += min_idx  # Adjust to original indices
    
    # Parabolic interpolation for sub-sample accuracy
    indices = torch.clamp(peak_indices, 1, frame_length-2)
    alpha = autocorr[torch.arange(num_frames), indices-1]
    beta = autocorr[torch.arange(num_frames), indices]
    gamma = autocorr[torch.arange(num_frames), indices+1]
    
    delta = 0.5 * (alpha - gamma) / (alpha - 2*beta + gamma + 1e-8)
    delta = torch.where((peak_indices > 0) & (peak_indices < frame_length-1), delta, 0.0)
    
    # Calculate final periods and pitches
    best_period = (peak_indices + delta) / sr
    pitch = torch.where(best_period > 0, 1.0 / best_period, 0.0)
    
    # Apply voicing threshold
    autocorr_0 = autocorr[:, 0]
    voiced = (peak_values / (autocorr_0 + 1e-8)) > threshold
    pitch = torch.where(voiced, pitch, 0.0)
    
    # Clamp valid frequencies
    pitch = torch.clamp(pitch, min_freq, max_freq)
    
    return pitch

def extract_single_pitch_value(
    audio_tensor: torch.Tensor,
    sr: int,
    min_freq: float = 75.0,
    max_freq: float = 600.0,
    frame_length: int = 400,
    hop_length: int = 160,
    threshold: float = 0.3
) -> float:
    """
    Calculates the average pitch of an audio tensor and normalizes it to 0-1 range.

    Args:
        audio_tensor: Input audio tensor (1D or 2D [channels, samples])
        sr: Sampling rate
        min_freq: Minimum detectable frequency (Hz)
        max_freq: Maximum detectable frequency (Hz)
        frame_length: Analysis frame length in samples
        hop_length: Hop size in samples
        threshold: Voicing threshold (0.0-1.0)

    Returns:
        A single float value representing the normalized average pitch (0.0-1.0).
    """
    pitch_tensor = calculate_pitch(
        audio_tensor, sr, min_freq, max_freq, frame_length, hop_length, threshold
    )

    # Calculate the average pitch across frames
    average_pitch = torch.mean(pitch_tensor)

    # Normalize to 0-1 range
    normalized_pitch = (average_pitch - min_freq) / (max_freq - min_freq)

    # Clamp to ensure it's strictly within 0-1 
    normalized_pitch = torch.clamp(normalized_pitch, 0.0, 1.0)

    return normalized_pitch.item() 

class Features:
    def __init__(self, device):
        self.eps = 1e-10
        self.device = torch.device(device if device is not None else "cuda" if torch.cuda.is_available() else "cpu")

    def scale_values(self, value: float) -> int:
        """
        Scale a value from [0,1] to [0,100] and round to nearest integer
        """
        return round(value * 100)

    def features_to_tokens(self, features: dict) -> list:
        """
        Convert features to token strings in format <|feature_value|>
        """
        return [f"<|{name}_{value}|>" for name, value in features.items()]
    
    def validate_audio(self, audio: torch.Tensor) -> bool:
        """
        Validate audio tensor before processing
        """
        if audio is None or not isinstance(audio, torch.Tensor):
            return False
        if audio.numel() == 0:  # Check if tensor is empty
            return False
        if torch.isnan(audio).any() or torch.isinf(audio).any():
            return False
        return True
    
    def get_default_features(self) -> dict:
        """
        Return default feature values when audio is invalid
        """
        return {
            'energy': 0,
            'spectral_centroid': 0,
            'pitch': 0
        }
    
    def extract_audio_features(self, audio: torch.Tensor, sr: int) -> dict:
        """
        Extract fast-to-compute features from audio segments.
        Each feature is normalized to [0, 1] range.
        
        Args:
            audio: Audio tensor of shape [channels, samples]
            sr: Sample rate
        
        Returns:
            Dictionary of features, each as a single float value
        """

        if not self.validate_audio(audio):
            return self.get_default_features()

        # Convert to mono if stereo
        if audio.dim() == 2 and audio.shape[0] > 1:
            audio = torch.mean(audio, dim=0, keepdim=True)
            
        # Get absolute values for amplitude calculations
        features = {}
        
        # RMS Energy (loudness) - normalized to [0, 1]
        features['energy'] = float(torch.sqrt(torch.mean(audio ** 2)))
        
        # Spectral Centroid - normalized to [0, 1]
        spec = torch.abs(torch.fft.rfft(audio))
        freqs = torch.linspace(0, sr/2, spec.shape[-1], device=self.device)
        spec_sum = torch.sum(spec) + self.eps
        centroid = torch.sum(freqs * spec.squeeze()) / spec_sum
        features['spectral_centroid'] = float(centroid / (sr/2)) 

        # Pitch - normalized to [0, 1]
        features['pitch'] = extract_single_pitch_value(audio, sr)

        for name, value in features.items():
            features[name] = self.scale_values(value)

        return features

class AudioProcessor:
    def __init__(self, config):
        self.device = torch.device(config.device if config.device is not None else "cuda" if torch.cuda.is_available() else "cpu")
        self.features = Features(config.device)
        self.audio_codec = DacInterface(config.device, config.audio_codec_path)

    def create_speaker_from_whisper(self, audio, whisper_model: str = "turbo", device = None):
        seconds = self.audio_codec.load_audio(audio).flatten().size(0) / self.audio_codec.sr
        if seconds > 20:
            raise ValueError("Speaker audio is longer than 20 seconds. Use a shorter clip for best results.")
        if seconds > 15:
            logger.warning("Speaker audio is longer than 15 seconds. For best results, consider using an audio clip up to 15 seconds.")
        
        data = transcribe_once_word_level(audio, whisper_model, device)
        text = text_normalizations(data['text'])
        words = []
        for s in data['segments']:
            words.extend([{'word': i['word'].strip(), 'start': float(i['start']), 'end': float(i['end'])} for i in s['words']])

        with open(audio, "rb") as f:
            audio_bytes = f.read()

        return self.create_speaker_from_dict({"audio": {"bytes": audio_bytes}, "text": text, "words": words})
    
    def create_speaker_from_dict(self, data: dict):
        audio = io.BytesIO(data['audio']['bytes'])
        audio = self.audio_codec.load_audio(audio).to(self.device)
        full_codes = self.audio_codec.encode(audio, verbose=True).tolist()[0]

        c1 = full_codes[0]
        c2 = full_codes[1]

        sr = self.audio_codec.sr
        text = data['text']
        words = data['words']

        tps = 75

        audio = audio.squeeze(0)
        global_features = self.features.extract_audio_features(audio, sr)

        start = None
        word_codes = []
        max_extension = 20

        for idx, i in enumerate(words):
            if start is None:
                start = max(0, int(i['start'] * tps) - max_extension)
            word = i['word'].strip()
            if idx == len(words) - 1:
                end = min(len(c1), int(i['end'] * tps) + max_extension)
            else:
                end = int(i['end'] * tps)

            word_c1 = c1[start:end]
            word_c2 = c2[start:end]

            word_audio = audio[:, int(i['start']*sr):int(i['end']*sr)]
            features = self.features.extract_audio_features(word_audio, sr)

            start = end

            word_codes.append({
                "word": word,
                "duration": round(len(word_c1) / tps, 2),
                "c1": word_c1,
                "c2": word_c2,
                "features": features
            })

        return {
            "text": text,
            "words": word_codes,
            "global_features": global_features
        }



import torchaudio
import torch
import os
import platform
from loguru import logger
from huggingface_hub import snapshot_download

from .model import WavEncoder, WavDecoder

class AudioCodec:
    def __init__(self, device: str = None, model_path: str = None, load_decoder: bool = True, load_encoder: bool = True):
        self.device = torch.device(device if device is not None else "cuda" if torch.cuda.is_available() else "cpu")
        self.cache_dir = self.get_cache_dir()
        if model_path is None:
            self.ensure_model_exists()
        else:
            if not os.path.isdir(model_path):
                raise ValueError(f"Model path {model_path} is not a directory. Please provide a valid directory path.")
            self.cache_dir = model_path

        self.encoder_path = os.path.join(self.cache_dir , 'encoder')
        if not os.path.isdir(self.encoder_path):
            raise ValueError(f"Encoder directory not found at {self.encoder_path}. The model path must contain an 'encoder' subdirectory.")
        
        self.decoder_path = os.path.join(self.cache_dir, 'decoder')
        if not os.path.isdir(self.decoder_path):
            raise ValueError(f"Decoder directory not found at {self.decoder_path}. The model path must contain a 'decoder' subdirectory.")
        
        self.sr = 24000
        self.bandwidth_id = torch.tensor([0])

        if load_decoder:
            self.load_decoder()
        
        if load_encoder:
            self.load_encoder()

    def get_cache_dir(self):
        return os.path.join(
            os.getenv('APPDATA') if platform.system() == "Windows" else os.path.join(os.path.expanduser("~"), ".cache"),
            "outeai", "tts", "wavtokenizer_75_token_interface")
    
    def ensure_model_exists(self):
        snapshot_download("OuteAI/wavtokenizer-large-75token-interface", local_dir=self.cache_dir)
    
    def load_encoder(self):
        self.encoder = WavEncoder.from_pretrained(self.encoder_path).to(self.device)
        self.encoder.eval()

    def load_decoder(self):
        self.decoder = WavDecoder.from_pretrained(self.decoder_path).to(self.device)
        self.decoder.eval()

    def convert_audio(self, wav: torch.Tensor, sr: int, target_sr: int, target_channels: int):
        assert wav.dim() >= 2, "Audio tensor must have at least 2 dimensions"
        assert wav.shape[-2] in [1, 2], "Audio must be mono or stereo."
        *shape, channels, length = wav.shape
        if target_channels == 1:
            wav = wav.mean(-2, keepdim=True)
        elif target_channels == 2:
            wav = wav.expand(*shape, target_channels, length)
        elif channels == 1:
            wav = wav.expand(target_channels, -1)
        else:
            raise RuntimeError(f"Impossible to convert from {channels} to {target_channels}")
        wav = torchaudio.transforms.Resample(sr, target_sr)(wav)
        return wav
    
    def convert_audio_tensor(self, audio: torch.Tensor, sr):
        return self.convert_audio(audio, sr, self.sr, 1)
    
    def load_audio(self, path):
        wav, sr = torchaudio.load(path)
        return self.convert_audio_tensor(wav, sr).to(self.device)

    def encode(self, audio: torch.Tensor):
        _, discrete_code = self.encoder(audio, bandwidth_id=self.bandwidth_id.to(self.device))
        return discrete_code

    def decode(self, codes):
        features = self.decoder.codes_to_features(codes)
        audio_out = self.decoder(features, bandwidth_id=self.bandwidth_id.to(self.device))
        return audio_out

    def save_audio(self, audio: torch.Tensor, path: str):
        torchaudio.save(path, audio.cpu(), sample_rate=self.sr, encoding='PCM_S', bits_per_sample=16)




    

       


import os
from loguru import logger

def create_speaker(
        audio_processor,
        audio_path: str,
        whisper_model: str = "turbo",
        whisper_device = None,
    ):
    if not os.path.exists(audio_path):
        raise FileNotFoundError(f"Audio file not found: {audio_path}")

    return audio_processor.create_speaker_from_whisper(
        audio=audio_path,
        whisper_model=whisper_model,
        device=whisper_device
    )


from dataclasses import dataclass, field
import torch
import os
from loguru import logger
from huggingface_hub import hf_hub_download
from pprint import pprint

from . import info
from ..utils import helpers

@dataclass
class SamplerConfig:
    temperature: float = 0.4
    repetition_penalty: float = 1.1
    repetition_range: int = 64
    top_k: int = 40
    top_p: float = 0.9
    min_p: float = 0.05
    mirostat_tau: int = 5
    mirostat_eta: float = 0.1
    mirostat: bool = False

@dataclass
class GenerationConfig:
    text: str
    voice_characteristics: str = None
    speaker: dict = None
    generation_type: info.GenerationType = info.GenerationType.CHUNKED
    max_batch_size: int = 16
    dac_decoding_chunk: int = 2048
    sampler_config: SamplerConfig = field(default_factory=SamplerConfig)
    max_length: int = 8192
    additional_gen_config: dict = field(default_factory=lambda: {})
    additional_dynamic_generator_config: dict = field(default_factory=lambda: {})
    server_host: str = "http://localhost:8080"

def get_compatible_dtype():
    """
    Returns the most compatible dtype for PyTorch based on the user's hardware:
    """
    if torch.cuda.is_available():
        if torch.cuda.is_bf16_supported():
            logger.info("BF16 support available. Using torch.bfloat16 data type.")
            return torch.bfloat16
        else:
            logger.info("BF16 support not available. Using torch.float16 data type.")
            return torch.float16
    else:
        logger.info("CUDA not available. Using torch.float32 data type.")
        return torch.float32
    
class LoadAutoModel:
    def __init__(self):
        self.cache = helpers.get_cache_dir()

    def init_model(
        self, 
        model: info.Models,
        backend: info.Backend,
        quantization: info.LlamaCppQuantization = None,
    ):
        """
        Automatically initializes and returns the path to the specified model.
        """
        model_dir = "OuteAI/" + model.value

        if backend == info.Backend.HF:
            return model_dir  # HF models are handled directly by Hugging Face
        
        elif backend == info.Backend.LLAMACPP:
            if quantization is None:
                raise ValueError("Quantization parameter is required for GGUF models.")
            model_dir += "-GGUF"
            filename = f"{model.value}-{quantization.value}.gguf"
            model_path = hf_hub_download(
                repo_id=model_dir,
                filename=filename,
                local_dir=os.path.join(self.cache, "gguf"),
                local_files_only=False
            )
            return model_path
        
        elif backend == info.Backend.EXL2:
            raise NotImplementedError("Automatic model loading for EXL2 models is not yet supported.")
        
        else:
            raise ValueError(f"Unsupported backend: {backend.value}")

class ModelConfig:
    def __init__(
        self,
        model_path: str = "OuteAI/Llama-OuteTTS-1.0-1B",
        tokenizer_path: str = "OuteAI/Llama-OuteTTS-1.0-1B",
        interface_version: info.InterfaceVersion = info.InterfaceVersion.V3,
        backend: info.Backend = info.Backend.HF,
        verbose: bool = False,
        device: str = None,
        dtype: torch.dtype = None,
        additional_model_config: dict = {},
        audio_codec_path: str = None,
        max_seq_length: int = 8192,
        n_gpu_layers: int = 0,
        exl2_cache_seq_multiply: int = 16,
        **kwargs
    ):
        self.model_path = model_path
        self.tokenizer_path = tokenizer_path
        self.interface_version = interface_version
        self.backend = backend 
        self.verbose = verbose
        self.device = device
        self.dtype = dtype
        self.additional_model_config = additional_model_config
        self.audio_codec_path = audio_codec_path
        self.max_seq_length = max_seq_length
        self.n_gpu_layers = n_gpu_layers
        self.exl2_cache_seq_multiply = exl2_cache_seq_multiply

        # Accept and set any extra keyword arguments as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.audio_processor = None
        self.audio_codec = None
        self.prompt_processor = None

        self._init_functions()

    def _init_functions(self):
        if self.interface_version == info.InterfaceVersion.V1:
            from ..version.v1.prompt_processor import PromptProcessor
            from ..wav_tokenizer.audio_codec import AudioCodec
            self.prompt_processor = PromptProcessor
            self.audio_codec = AudioCodec
        elif self.interface_version == info.InterfaceVersion.V2:
            from ..version.v2.prompt_processor import PromptProcessor
            from ..wav_tokenizer.audio_codec import AudioCodec
            self.prompt_processor = PromptProcessor
            self.audio_codec = AudioCodec
        elif self.interface_version == info.InterfaceVersion.V3:
            from ..version.v3.prompt_processor import PromptProcessor
            from ..version.v3.audio_processor import AudioProcessor
            from ..dac.interface import DacInterface
            self.prompt_processor = PromptProcessor
            self.audio_processor = AudioProcessor
            self.audio_codec = DacInterface
        else:
            raise ValueError(f"Unsupported interface version: {self.interface_version}")
    
    @classmethod
    def auto_config(cls, model: info.Models, backend: info.Backend, quantization: info.LlamaCppQuantization = None):
        logger.info(f"Initializing model configuration for {model.value} model with {backend.value} backend.")
        if model != info.Models.VERSION_1_0_SIZE_1B and model != info.Models.VERSION_1_0_SIZE_0_6B:
            raise ValueError("Only OuteTTS 1.0 model supported for auto configuration.")
        
        model_path = LoadAutoModel().init_model(model, backend, quantization)
        model_info = info.MODEL_INFO[model]

        logger.info(f"Model path: {model_path}")
        config = {
            "model_path": model_path,
            "tokenizer_path": "OuteAI/" + model.value,
            "interface_version": model_info["interface_version"],
            "backend": backend,
            "verbose": False,
            "device": None,
            "dtype": None,
            "additional_model_config": {},
            "audio_codec_path": None,
            "max_seq_length": model_info["max_seq_length"],
            "n_gpu_layers": 0,
        }

        if backend == info.Backend.HF:
            config["dtype"] = get_compatible_dtype()
            try:
                from flash_attn import flash_attn_qkvpacked_func, flash_attn_func
                config["additional_model_config"]["attn_implementation"] = "flash_attention_2"
                logger.success("Flash attention available. Using flash_attention_2 implementation.")
            except:
                logger.warning("Flash attention 2 not available. Using default attention implementation.\nFor faster inference on supported hardware, consider installing FlashAttention using:\npip install flash-attn --no-build-isolation")
        elif backend == info.Backend.LLAMACPP:
            config["n_gpu_layers"] = 99
            logger.info("LLAMA.CPP backend selected. Offloading all layers to GPU.")

        else:
            raise NotImplementedError(f"Automatic model loading for {backend.value} is not yet supported.")
        
        logger.info("Using config:")
        pprint(config, indent=2)
        
        return cls(**config)





        


from transformers import AutoTokenizer

from .tokens import SpecialTokens
from .text_processor import TextProcessor

class PromptProcessor:
    def __init__(self, tokenizer_path: str):
        if tokenizer_path:
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)

        self.special_tokens = SpecialTokens()

        self.text_prompt = "{bos}\n{text_start}{words}{text_end}\n"

        if tokenizer_path:
            self.map_audio_tokens = self.get_audio_token_map()

        self.text_processor = TextProcessor()

    def get_audio_token_map(self) -> dict:
        return {
            self.tokenizer.encode(self.special_tokens.audio_code.format(i), add_special_tokens=False)[0]: i
            for i in range(4100)
        }

    def create_audio_prompt(self, processed, speaker):
        word_data = speaker["words"]
        words = []
        idx = 0  # Separate index for words

        for i in processed:
            if "<|emotion_start|>" in i["word"]:
                words.append(i["word"])
                continue

            s = word_data[idx]
            if s["word"] != i["word"]:
                raise ValueError(
                    f"Word mismatch at index {idx}:\n"
                    f"Expected word: '{s['word']}'\n"
                    f"Found word: '{i['word']}'"
                )

            word = s["word"]
            if i.get("before", []):
                word = "".join(i["before"]) + word
            if i.get("after", []):
                word += "".join(i["after"])

            tokens = "".join([self.special_tokens.audio_code.format(c) for c in s["codes"]])
            word = f"{word}{self.special_tokens.time.format(s['duration'])}{tokens}"
            
            words.append(word)
            idx += 1  # Only increment word index for non-emotion words
        
        words = (self.special_tokens.space+"\n").join(words)
        return words

    def get_completion_prompt(self, text: str, emotion: str = None, speaker: dict = None) -> str:
        words = self.text_processor.process_text(text)
        words = self.text_processor.get_text(words)
        if speaker is not None:
            speaker_text = self.text_processor.process_text(speaker['text'])
            audio = self.create_audio_prompt(speaker_text, speaker)
            words = self.text_processor.get_text(speaker_text) + self.special_tokens.space + words

        prompt = self.text_prompt.format(
            bos=self.special_tokens.bos, 
            text_start=self.special_tokens.text_start,
            words=words, 
            text_end=self.special_tokens.text_end,
        )
        if emotion is not None:
            prompt += f"{self.special_tokens.voice_characteristic_start}{emotion}{self.special_tokens.voice_characteristic_end}\n"
        prompt += self.special_tokens.audio_start + "\n"

        if speaker is not None:
            prompt += audio
            prompt += self.special_tokens.space + "\n"

        return prompt

    def get_training_prompt(self, speaker: dict) -> str:
        emotion = speaker.get("emotion", None)
        speaker_text = self.text_processor.process_text(speaker['text'])
        audio = self.create_audio_prompt(speaker_text, speaker)
        words = self.text_processor.get_text(speaker_text)
        prompt = self.text_prompt.format(
            bos=self.special_tokens.bos, 
            text_start=self.special_tokens.text_start,
            words=words, 
            text_end=self.special_tokens.text_end,
        )
        if emotion is not None:
            prompt += f"{self.special_tokens.voice_characteristic_start}{emotion}{self.special_tokens.voice_characteristic_end}\n"
        prompt += self.special_tokens.audio_start + "\n"

        prompt += audio
        prompt += "\n" + self.special_tokens.audio_end + "\n" + self.special_tokens.eos + "\n"
        return prompt

    def extract_audio_from_tokens(self, tokens: list[int]) -> list[int]:
        return [self.map_audio_tokens[i] for i in tokens if i in self.map_audio_tokens]



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r', encoding='utf-8') as fh:
    install_requires = fh.read().splitlines()

setup(
    name='outetts',
    version='0.4.4',
    packages=find_packages(),
    install_requires=install_requires,
    author='OuteAI',
    description='OuteAI Text-to-Speech (TTS)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/edwko/OuteTTS',
    package_data={
        "outetts.version.v3": ["default_speakers/json/*"],
        "outetts.anyascii": ["_data/*"],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)



import torch
from loguru import logger
import os
import json
import polars as pl
import re

from ..models.config import ModelConfig, GenerationConfig, info
from ..models.hf_model import HFModel
from ..models.gguf_model import GGUFModel
from ..models.exl2_model import EXL2Model, EXL2ModelAsync
from ..models.vllm_model import VLLMModelBatch
from ..models.llamacpp_server import LlamaCPPServerModel, LlamaCPPServerAsyncModel
from ..utils.chunking import chunk_text
from ..utils import preprocessing
from .playback import ModelOutput

class InterfaceHF:
    def __init__(
        self,
        config: ModelConfig,
    ) -> None:
        self.config = config
        self.verbose = config.verbose
        if config.interface_version == info.InterfaceVersion.V3:
            self.audio_processor = config.audio_processor(config)
            self.audio_codec = self.audio_processor.audio_codec
        else:
            self.audio_codec = config.audio_codec(config.device, config.audio_codec_path)

        self.prompt_processor = config.prompt_processor(config.tokenizer_path)
        self.model = self.get_model()
        
    def get_model(self):
        return HFModel(
            self.config.model_path, 
            self.config.device, 
            self.config.dtype, 
            self.config.additional_model_config
        )

    def _prepare_prompt(self, prompt: str):
        return self.prompt_processor.tokenizer.encode(
            prompt,
            add_special_tokens=False,
            return_tensors="pt"
        ).to(self.model.device)
    
    def prepare_prompt(self, text: str, speaker: dict = None):
        prompt = self.prompt_processor.get_completion_prompt(text, speaker)
        return self._prepare_prompt(prompt)
    
    def get_audio(self, tokens):
        output = self.prompt_processor.extract_audio_from_tokens(tokens)
        if not output:
            logger.error("No audio tokens found in the output")
            return None

        return self.audio_codec.decode(
            torch.tensor([output], dtype=torch.int64).to(self.audio_codec.device)
        )

    def create_speaker(
            self,
            audio_path: str,
            transcript: str = None,
            whisper_model: str = "turbo",
            whisper_device = None,
        ):
        if transcript and self.config.interface_version == info.InterfaceVersion.V3:
            logger.warning("V3 Interface models do not support speaker creation from transcripts."
                            "Whisper will be used to extract speaker features from the audio file."
                            "This is left for older interface compatibility and may be removed in future versions.")
            
        if self.config.interface_version == info.InterfaceVersion.V3:
            from ..version.v3.utils import create_speaker
            speaker = create_speaker(
                audio_processor=self.audio_processor,
                audio_path=audio_path,
                whisper_model=whisper_model,
                whisper_device=whisper_device
            )
        elif self.config.interface_version == info.InterfaceVersion.V2:
            from ..version.v2.utils import create_speaker
            speaker = create_speaker(
                device=self.config.device,
                audio_codec=self.audio_codec,
                audio_path=audio_path,
                transcript=transcript,
                whisper_model=whisper_model,
                whisper_device=whisper_device
            )
        elif self.config.interface_version == info.InterfaceVersion.V1:
            from ..version.v1.utils import create_speaker
            speaker = create_speaker(
                device=self.config.device,
                audio_codec=self.audio_codec,
                audio_path=audio_path,
                transcript=transcript,
                whisper_model=whisper_model,
                whisper_device=whisper_device
            )
        else:
            raise ValueError(f"Unsupported interface version: {self.config.interface_version}") 
        speaker['interface_version'] = self.config.interface_version.value
        return speaker

    def save_speaker(self, speaker: dict, path: str):
        speaker['interface_version'] = self.config.interface_version.value
        if not path.endswith(".json"):
            path += ".json"
        with open(path, "w") as f: 
            json.dump(speaker, f, indent=2)
        logger.info(f"Speaker saved to: {path}")

    def load_speaker(self, path: str):
        with open(path, "r") as f: 
            speaker = json.load(f)

        version = speaker.get('interface_version', 2)

        # For backwards compatibility
        current_version = self.config.interface_version.value 
        if version == 2 and current_version <= 2:
            return speaker

        if current_version != version:
            raise ValueError(f"Speaker interface version mismatch: {version} != {self.config.interface_version.value}")
        
        return speaker
    
    def decode_and_save_speaker(self, speaker: dict, path: str):
        if self.config.interface_version != info.InterfaceVersion.V3:
            raise ValueError("Speaker decoding is only supported for InterfaceVersion.V3")
        c1, c2 = [], []
        for i in speaker["words"]:
            c1.extend(i["c1"])
            c2.extend(i["c2"])
        ModelOutput(self.audio_codec.decode(
            torch.tensor([[c1,c2]], dtype=torch.int64).to(self.audio_codec.device)
        ), og_sr=self.audio_codec.sr).save(path)

    def print_default_speakers(self):
        if self.config.interface_version == info.InterfaceVersion.V3:
            # _BASE_DIR = os.path.dirname(__file__)
            # _DEFAULT_SPEAKERS_DIR = os.path.join(_BASE_DIR, "v3/default_speakers/parquet/speakers.parquet")
            # if not os.path.exists(_DEFAULT_SPEAKERS_DIR):
            #     raise FileNotFoundError(f"Default speakers file not found: {_DEFAULT_SPEAKERS_DIR}")
            # _DEFAULT_SPEAKERS = {i['speaker']: i for i in pl.read_parquet(_DEFAULT_SPEAKERS_DIR).to_dicts()}
            logger.info(f"Available default speakers v3: {['en-female-1-neutral']}")
        else:
            logger.warning("Default speakers are not supported for this interface version.")

    def load_default_speaker(self, name: str):
        name = name.lower().strip()
        if self.config.interface_version == info.InterfaceVersion.V3:
            _BASE_DIR = os.path.dirname(__file__)
            _DEFAULT_SPEAKER_DIR = os.path.join(_BASE_DIR, "v3/default_speakers/json/en-female-1-neutral.json")
            # _DEFAULT_SPEAKERS_DIR = os.path.join(_BASE_DIR, "v3/default_speakers/parquet/speakers.parquet")
            # _DEFAULT_SPEAKERS_DIR = os.path.join(_BASE_DIR, "v3/default_speakers/parquet/speakers.parquet")
            # if not os.path.exists(_DEFAULT_SPEAKERS_DIR):
            #     raise FileNotFoundError(f"Default speakers file not found: {_DEFAULT_SPEAKERS_DIR}")
            # _DEFAULT_SPEAKERS = {i['speaker']: i for i in pl.read_parquet(_DEFAULT_SPEAKERS_DIR).to_dicts()}
            with open(_DEFAULT_SPEAKER_DIR, "r") as f:
                speaker = json.load(f)
            _DEFAULT_SPEAKERS = {"en-female-1-neutral": speaker}
            if name not in _DEFAULT_SPEAKERS:
                raise ValueError(f"Speaker {name} not found {list(_DEFAULT_SPEAKERS.keys())}")
            return _DEFAULT_SPEAKERS.get(name, {})
        else:
            raise ValueError("Default speakers are not supported for this interface version.")

    def check_generation_max_length(self, max_length):
        if max_length is None:
            raise ValueError("max_length must be specified.")
        if max_length > self.config.max_seq_length:
            raise ValueError(f"Requested max_length ({max_length}) exceeds the current max_seq_length ({self.config.max_seq_length}).")

    def _generate(self, input_ids, config: GenerationConfig):
        output = self.model.generate(
            input_ids=input_ids,
            config=config
        )
        return output[input_ids.size()[-1]:]
    
    def guided_words_generation(self, config: GenerationConfig):
        text_chunks = chunk_text(config.text)
        chunk_size = len(text_chunks)
        word_token = self.prompt_processor.tokenizer.encode(
            self.prompt_processor.special_tokens.word_start, add_special_tokens=False
        )[0]
        end_token = self.prompt_processor.tokenizer.encode(
            self.prompt_processor.special_tokens.audio_end, add_special_tokens=False
        )[0]
        logger.info(f"Created: {chunk_size} text chunks")
        all_outputs = []

        def create_insert(word):
            insert = [word_token] 
            insert.extend(self.prompt_processor.tokenizer.encode(
                word + self.prompt_processor.special_tokens.features, add_special_tokens=False))
            return insert
        
        def cat_inputs(input_ids, output):
            if isinstance(input_ids, torch.Tensor):
                return torch.cat(
                    [input_ids, torch.tensor([output], dtype=torch.int64).to(input_ids.device)], dim=1
                )
            elif isinstance(input_ids, list):
                input_ids.extend(output)
                return input_ids
            elif isinstance(input_ids, str):
                return input_ids + self.prompt_processor.tokenizer.decode(output, skip_special_tokens=False)
            else:
                raise ValueError("Invalid input_ids type")

        for i, chunk in enumerate(text_chunks):
            logger.info(f"Proccessing: Chunk {i+1} / {chunk_size}")
            words = preprocessing.get_words(chunk)
            # Initialize first word
            first_word = words.pop(0)
            print(f"\nInserting first word: {repr(first_word)}")
            input_ids = self.prepare_prompt(chunk + first_word + self.prompt_processor.special_tokens.features, config.speaker)
            output = []
            break_next = False
            while True:
                for token in self.model._generate_stream(input_ids, config):
                    if token == word_token or token == end_token:
                        if not words:
                            break
                        insert = create_insert(words.pop(0))
                        output.extend(insert)
                        all_outputs.extend(insert)
                        print(f"\nInserting: {repr(self.prompt_processor.tokenizer.decode(insert, skip_special_tokens=False))}")
                        input_ids = cat_inputs(input_ids, output)
                        output = []
                        break
                    else:
                        output.append(token)
                        all_outputs.append(token)

                if not words:
                    if break_next:
                        break
                    break_next = True

        return all_outputs

    def chunk_generation(self, config: GenerationConfig):
        text_chunks = chunk_text(config.text)
        audio_chunks = []
        chunk_size = len(text_chunks)

        logger.info(f"Created: {chunk_size} text chunks")
        for i, chunk in enumerate(text_chunks):
            logger.info(f"Proccessing: Chunk {i+1} / {chunk_size}")

            input_ids = self.prepare_prompt(chunk, config.speaker)

            output = self._generate(input_ids, config)
            audio_chunks.extend(output)

        return audio_chunks
    
    def regular_generation(self, config: GenerationConfig):
        input_ids = self.prepare_prompt(config.text, config.speaker)
        return self._generate(input_ids, config)
    
    def batch_generation(self, config: GenerationConfig):
        text_chunks = chunk_text(config.text)
        logger.info(f"Total text batches created: {len(text_chunks)}")

        input_ids = [self.prepare_prompt(i, config.speaker) for i in text_chunks]
        outputs = []

        output = self._generate_batch(input_ids, config)
        outputs.extend([i['text'] for i in output])

        print("\n")
        logger.success("Batch generation finished!")

        c1 = []
        c2 = []
        for i in outputs:
            codebook1 = [int(x) for x in re.findall(r'<\|c1_(\d+)\|>', i)]
            codebook2 = [int(x) for x in re.findall(r'<\|c2_(\d+)\|>', i)]

            if not codebook1 or not codebook2:
                logger.error("Codebook is empty. The audio chunk might be missing segment, which could result in incomplete audio when combined. Skipping...")
                continue
                
            min_len = min(len(codebook1), len(codebook2))
            c1.extend(codebook1[:min_len])
            c2.extend(codebook2[:min_len])
        
        logger.info("Decoding audio")

        outputs = self.audio_codec.decode(
            torch.tensor([[c1,c2]], dtype=torch.int64),
            verbose=True,
            chunk_length=config.dac_decoding_chunk
        )

        return outputs

    def _batch_check(self, config: GenerationConfig):
        if config.generation_type != info.GenerationType.BATCH:
            logger.info("vLLM, EXL2 Async, Llama.cpp async server only support GenerationType.BATCH.")
            logger.info(f"Generation type '{config.generation_type.value}' is not supported. Switching to 'BATCH' mode.")
            return info.GenerationType.BATCH
    
    def generate(self, config: GenerationConfig) -> ModelOutput:
        self.check_generation_max_length(config.max_length)
        if config.text is None:
            raise ValueError("text can not be empty!")

        if self.config.backend in [info.Backend.VLLM, info.Backend.EXL2ASYNC, info.Backend.LLAMACPP_ASYNC_SERVER]:
            _batch_check = self._batch_check(config)
            if _batch_check is not None:
                config.generation_type = _batch_check
        
        if config.generation_type == info.GenerationType.CHUNKED:
            output = self.chunk_generation(config)
        elif config.generation_type == info.GenerationType.GUIDED_WORDS:
            logger.warning("Guided words generation is experimental and may not work as expected.")
            if self.config.interface_version != info.InterfaceVersion.V3:
                raise ValueError("Guided words generation is only supported for InterfaceVersion.V3")
            if self.config.backend == info.Backend.HF or self.config.backend == info.Backend.EXL2:
                raise ValueError("Guided words generation supports only llama.cpp backend.")
            output = self.guided_words_generation(config)
        elif config.generation_type == info.GenerationType.REGULAR:
            logger.info("Using regular generation, consider using chunked generation for long texts.")
            output = self.regular_generation(config)
        elif config.generation_type == info.GenerationType.BATCH:
            if self.config.backend == info.Backend.HF or self.config.backend == info.Backend.LLAMACPP or self.config.backend == info.Backend.LLAMACPP_SERVER:
                raise ValueError("Batch generation supports only GenerationType.VLLM, GenerationType.EXL2ASYNC, info.Backend.LLAMACPP_ASYNC_SERVER backends.")
            audio = self.batch_generation(config)
            return ModelOutput(audio, self.audio_codec.sr)
        else:
            raise ValueError(f"Unsupported generation type: {config.generation_type}")
        
        audio = self.get_audio(output)
        return ModelOutput(audio, self.audio_codec.sr)

class InterfaceLLAMACPP(InterfaceHF):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        return GGUFModel(
            model_path=self.config.model_path,
            n_gpu_layers=self.config.n_gpu_layers,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config
        )

    def _generate(self, input_ids, config):
        return self.model.generate(
            input_ids=input_ids,
            config=config,
        )

    def _prepare_prompt(self, prompt: str):
        return self.prompt_processor.tokenizer.encode(prompt, add_special_tokens=False)

class InterfaceEXL2(InterfaceHF):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        return EXL2Model(
            model_path=self.config.model_path,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config,
        )

    def _prepare_prompt(self, prompt: str):
        return prompt
    
    def _generate(self, input_ids, config):
        return self.model.generate(
            input_ids=input_ids,
            config=config,
        )
    
class InterfaceEXL2Async(InterfaceHF):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        if "exl2_cache_size" not in self.config.additional_model_config:
            self.config.additional_model_config["exl2_cache_size"] = self.config.max_seq_length * self.config.exl2_cache_seq_multiply
        return EXL2ModelAsync(
            model_path=self.config.model_path,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config,
        )

    def _prepare_prompt(self, prompt: str):
        return prompt
    
    def _generate_batch(self, input_ids, config):
        return self.model.generate_batch(
            input_ids=input_ids,
            config=config,
        )

class InterfaceVLLMBatch(InterfaceHF):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        return VLLMModelBatch(
            model_path=self.config.model_path,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config,
        )

    def _prepare_prompt(self, prompt: str):
        return prompt
    
    def _generate_batch(self, input_ids, config):
        return self.model.generate_batch(
            input_ids=input_ids,
            config=config,
        )

class InterfaceLlamaCPPServerAsyncBatch(InterfaceHF):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        return LlamaCPPServerAsyncModel(
            model_path=self.config.model_path,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config,
        )

    def _prepare_prompt(self, prompt: str):
        return prompt
    
    def _generate_batch(self, input_ids, config):
        return self.model.generate_batch(
            input_ids=input_ids,
            config=config,
        )

class InterfaceLlamaCPPServer(InterfaceHF):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.config = config

    def get_model(self):
        return LlamaCPPServerModel(
            model_path=self.config.model_path,
            max_seq_length=self.config.max_seq_length,
            additional_model_config=self.config.additional_model_config,
            tokenizer=self.prompt_processor.tokenizer
        )

    def _prepare_prompt(self, prompt: str):
        return prompt
    
    def _generate(self, input_ids, config):
        return self.model.generate(
            input_ids=input_ids,
            config=config,
        )
    


from enum import Enum

class Backend(Enum):
    HF = "hf"
    LLAMACPP = "llamacpp"
    EXL2 = "exl2"
    EXL2ASYNC = "exl2async"
    VLLM = "vllm"
    LLAMACPP_SERVER = "llamacpp_server"
    LLAMACPP_ASYNC_SERVER = "llamacpp_async_server"

class InterfaceVersion(Enum):
    V1 = 1
    V2 = 2
    V3 = 3

class Models(Enum):
    VERSION_0_1_SIZE_350M = "OuteTTS-0.1-350M"
    VERSION_0_2_SIZE_500M = "OuteTTS-0.2-500M"
    VERSION_0_3_SIZE_500M = "OuteTTS-0.3-500M"
    VERSION_0_3_SIZE_1B = "OuteTTS-0.3-1B"
    VERSION_1_0_SIZE_1B = "Llama-OuteTTS-1.0-1B"
    VERSION_1_0_SIZE_0_6B = "OuteTTS-1.0-0.6B"

class GenerationType(Enum):
    REGULAR = "regular"
    CHUNKED = "chunked"
    GUIDED_WORDS = "guided_words"
    STREAM = "stream"
    BATCH = "batch"

class LlamaCppQuantization(Enum):
    FP16 = "FP16"
    Q8_0 = "Q8_0"
    Q6_K = "Q6_K"
    Q5_K_S = "Q5_K_S"
    Q5_K_M = "Q5_K_M"
    Q5_1 = "Q5_1"
    Q5_0 = "Q5_0"
    Q4_K_S = "Q4_K_S"
    Q4_K_M = "Q4_K_M"
    Q4_1 = "Q4_1"
    Q4_0 = "Q4_0"
    Q3_K_S = "Q3_K_S"
    Q3_K_M = "Q3_K_M"
    Q3_K_L = "Q3_K_L"
    Q2_K = "Q2_K"

MODEL_INFO = {
    Models.VERSION_0_1_SIZE_350M: {
        "max_seq_length": 4096,
        "interface_version": InterfaceVersion.V1, 
    },
    Models.VERSION_0_2_SIZE_500M: {
        "max_seq_length": 4096,
        "interface_version": InterfaceVersion.V2,
    },
    Models.VERSION_0_3_SIZE_500M: {
        "max_seq_length": 4096,
        "interface_version": InterfaceVersion.V2,
    },
    Models.VERSION_0_3_SIZE_1B: {
        "max_seq_length": 4096,
        "interface_version": InterfaceVersion.V2,
    },
    Models.VERSION_1_0_SIZE_1B: {
        "max_seq_length": 8192,
        "interface_version": InterfaceVersion.V3,
    },
    Models.VERSION_1_0_SIZE_0_6B: {
        "max_seq_length": 8192,
        "interface_version": InterfaceVersion.V3,
    },
}


from dataclasses import dataclass, asdict
from typing import Dict

@dataclass
class SpecialTokens:
    """
    Dataclass containing special tokens used for text and audio processing.
    """
    bos: str = "<|im_start|>"
    eos: str = "<|im_end|>"
    c1: str = "<|c1_{}|>"
    c2: str = "<|c2_{}|>"
    text_start: str = "<|text_start|>"
    text_end: str = "<|text_end|>"
    voice_characteristic_start: str = "<|voice_characteristic_start|>"
    voice_characteristic_end: str = "<|voice_characteristic_end|>"
    emotion_start: str = "<|emotion_start|>"
    emotion_end: str = "<|emotion_end|>"
    audio_start: str = "<|audio_start|>"
    audio_end: str = "<|audio_end|>"
    time: str = "<|t_{:.2f}|>"
    code: str = "<|code|>"
    energy: str = "<|energy_{}|>"
    spectral_centroid: str = "<|spectral_centroid_{}|>"
    pitch: str = "<|pitch_{}|>"
    word_start: str = "<|word_start|>"
    word_end: str = "<|word_end|>"
    features: str = "<|features|>"
    global_features_start: str = "<|global_features_start|>"
    global_features_end: str = "<|global_features_end|>"

    def to_dict(self) -> Dict[str, str]:
        """Convert the dataclass instance to a dictionary using asdict."""
        return asdict(self)


from loguru import logger
import time
from tqdm import tqdm
from tqdm.auto import tqdm as atqdm
import asyncio
import torch
import uuid

from .config import GenerationConfig
from .info import GenerationType

import logging
from typing import Tuple

try:
    from vllm.engine.arg_utils import AsyncEngineArgs
    from vllm.engine.async_llm_engine import AsyncLLMEngine
    from vllm.sampling_params import SamplingParams
    from vllm.utils import random_uuid
    import vllm.model_executor.layers.utils
    _VLLM_AVAILABLE = True
except:
    _VLLM_AVAILABLE = False

def get_recent_tokens(prompt_tokens: torch.Tensor, 
                      output_tokens: torch.Tensor,
                      window_size: int,
                      vocab_size: int) -> torch.Tensor:

    num_seqs = prompt_tokens.shape[0]
    device = prompt_tokens.device
    
    recent_tokens = torch.full((num_seqs, window_size), 
                               vocab_size,
                               dtype=torch.long, 
                               device=device)
    
    for seq_idx in range(num_seqs):
        valid_prompt = prompt_tokens[seq_idx][prompt_tokens[seq_idx] != vocab_size]
        
        valid_output = output_tokens[seq_idx][output_tokens[seq_idx] != vocab_size]
        
        all_tokens = torch.cat([valid_prompt, valid_output])
        
        token_count = min(window_size, all_tokens.shape[0])
        if token_count > 0:
            recent_tokens[seq_idx, -token_count:] = all_tokens[-token_count:]
    
    return recent_tokens

def get_token_bin_counts_and_mask(
    tokens: torch.Tensor,
    vocab_size: int,
    num_seqs: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    bin_counts = torch.zeros((num_seqs, vocab_size + 1),
                             dtype=torch.long,
                             device=tokens.device)
    bin_counts.scatter_add_(1, tokens, torch.ones_like(tokens))
    bin_counts = bin_counts[:, :vocab_size]
    mask = bin_counts > 0

    return bin_counts, mask

def apply_penalties(logits: torch.Tensor, prompt_tokens_tensor: torch.Tensor,
                    output_tokens_tensor: torch.Tensor,
                    presence_penalties: torch.Tensor,
                    frequency_penalties: torch.Tensor,
                    repetition_penalties: torch.Tensor,
                    ) -> torch.Tensor:
    rep_window_size = 64
    num_seqs, vocab_size = logits.shape
    
    _, prompt_mask = get_token_bin_counts_and_mask(prompt_tokens_tensor,
                                                  vocab_size, num_seqs)
    output_bin_counts, output_mask = get_token_bin_counts_and_mask(
        output_tokens_tensor, vocab_size, num_seqs)
    
    if rep_window_size is not None:
        recent_tokens = get_recent_tokens(prompt_tokens_tensor, 
                                         output_tokens_tensor, 
                                         rep_window_size,
                                         vocab_size)
        
        _, rep_mask = get_token_bin_counts_and_mask(recent_tokens,
                                                  vocab_size, num_seqs)
    else:
        rep_mask = prompt_mask | output_mask
    
    repetition_penalties = repetition_penalties.unsqueeze(dim=1).repeat(
        1, vocab_size)
    logits[logits > 0] /= torch.where(rep_mask,
                                     repetition_penalties, 1.0)[logits > 0]
    logits[logits <= 0] *= torch.where(rep_mask,
                                      repetition_penalties, 1.0)[logits <= 0]

    logits -= frequency_penalties.unsqueeze(dim=1) * output_bin_counts
    logits -= presence_penalties.unsqueeze(dim=1) * output_mask
    return logits

if _VLLM_AVAILABLE:
    vllm.model_executor.layers.utils.apply_penalties = apply_penalties

class VLLMModelBatch:
    def __init__(
            self,
            model_path: str,
            max_seq_length: int,
            additional_model_config: dict = {},
    ) -> None:

        if not _VLLM_AVAILABLE:
            raise ImportError(
                "VLLM python module not found."
                "To use the VLLM you must install it manually."
            )
        
        model_config = AsyncEngineArgs(
            model=model_path,
            tokenizer=model_path,
            tokenizer_mode="auto",
            trust_remote_code=True, 
            tensor_parallel_size=additional_model_config.get("vllm_tensor_parallel_size", 1),
            max_model_len=8192, 
            gpu_memory_utilization=additional_model_config.get("vllm_gpu_memory_utilization", 0.9),
            disable_log_stats=True,
        )
        self.model = AsyncLLMEngine.from_engine_args(model_config)

        logging.getLogger("vllm").setLevel(logging.WARNING)

        self.loop = asyncio.new_event_loop()

        self.sem = None

    async def _generate(
            self, 
            input_ids: str, 
            config: GenerationConfig, 
            request_id,
            pbar: atqdm,
            lock: asyncio.Lock
        ):

        async with self.sem:

            sampling_params = SamplingParams(
                temperature=config.sampler_config.temperature,
                repetition_penalty=config.sampler_config.repetition_penalty,
                top_k=config.sampler_config.top_k,
                top_p=config.sampler_config.top_p,
                min_p=config.sampler_config.min_p,
                frequency_penalty=0.0,
                presence_penalty=0.0,
                max_tokens=config.max_length,
                skip_special_tokens=False
            )

            results_generator = self.model.generate(input_ids, sampling_params, uuid.uuid4().hex)
            tokens = []
            output_text = ""

            async for result in results_generator:
                current_output_text = result.outputs[0].text
                delta_text = current_output_text[len(output_text):]
                output_text = current_output_text

                if not delta_text:
                    await asyncio.sleep(0)
                    continue
                tokens.extend(delta_text)

                async with lock:
                    pbar.update(1)
                    speed = pbar.format_dict.get("rate", 0)
                    rtf = round(1 / (speed / 150), 4) if speed else 0
                    pbar.set_postfix({"RTF": rtf})

            print("\n")
            logger.success(f"Batch {request_id} finished.\n")

            return {"text": "".join(tokens), "id": request_id}
    
    async def _generate_batch_async(self, input_ids: list[str], config: GenerationConfig):
        """Helper coroutine to run multiple _generate tasks in parallel and gather results in order."""
        pbar = atqdm(
            total=None,
        )
        lock = asyncio.Lock()
        self.sem = asyncio.Semaphore(config.max_batch_size)

        tasks = [
            asyncio.create_task(
                self._generate(i, config, idx, pbar, lock)
            )
            for idx, i in enumerate(input_ids)
        ]
        return await asyncio.gather(*tasks)

    def generate_batch(self, input_ids: list[str], config: GenerationConfig):
        output = self.loop.run_until_complete(
            self._generate_batch_async(input_ids, config)
        )
        torch.cuda.empty_cache()
        return output

        


import torch
from loguru import logger

from .alignment import CTCForcedAlignment
from ...whisper import transcribe

def clean_dashes(text: str) -> str:
    dashes = ['—', '–', '-']
    for dash in dashes:
        text = text.replace(dash, ' ')
    return text

def create_speaker(
        device,
        audio_codec,
        audio_path: str,
        transcript: str = None,
        whisper_model: str = "turbo",
        whisper_device = None,
    ):
    if transcript is None:
        logger.info("Transcription not provided, transcribing audio with whisper.")
        transcript = transcribe.transcribe_once(
            audio_path=audio_path,
            model=whisper_model,
            device=whisper_device
        )
    if not transcript: 
        raise ValueError("Transcript text is empty")

    ctc = CTCForcedAlignment(device)

    words = ctc.align(audio_path, transcript)
    full_codes = audio_codec.encode(
        audio_codec.convert_audio_tensor(
            audio=torch.cat([i["audio"] for i in words], dim=1),
            sr=ctc.sample_rate
        ).to(audio_codec.device)
    ).tolist()

    data = []
    start = 0
    for i in words:
        end = int(round((i["x1"] / ctc.sample_rate) * 75))
        word_tokens = full_codes[0][0][start:end]
        start = end
        if not word_tokens:
            word_tokens = [1]

        data.append({
            "word": i["word"],
            "duration": round(len(word_tokens) / 75, 2),
            "codes": word_tokens
        })

    ctc.free()
    del ctc

    return {
        "text": transcript,
        "words": data,
    }